"""" Okushiri Tsunami Test Case (101 hr)

The numerical case is described in "Shallow Water and Navier-Stokes SPH-like
numerical modelling of rapidly varying free-surface flows", Renato Vacondio,
Doctoral dissertation, University of Parma, pp 104-112 (2010).

The experimental data can be obtained from the below url
http://isec.nacse.org/workshop/2004_cornell/bmark2.html

"""
import os

# Numpy
from numpy import (ones_like, zeros, zeros_like, mgrid, ravel, loadtxt, arange,
                   sqrt, concatenate, where, array, savez)

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.solver.utils import load

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        GatherDensityEvalNextIteration, NonDimensionalDensityResidual,
        UpdateSmoothingLength, CheckConvergenceDensityResidual, SWEOS,
        SWEIntegrator, SWEStep, CorrectionFactorVariableSmoothingLength,
        ParticleAcceleration, FluidBottomElevation, FluidBottomGradient,
        FluidBottomCurvature, GradientCorrectionPreStep, GradientCorrection,
        BedGradient, BedCurvature, SWEInlet, SWEInletOutletStep,
        RemoveFluidParticlesWithNoNeighbors, RemoveParticlesWithZeroAlpha,
        RemoveCloseParticlesAtOpenBoundary, BoundaryInnerReimannStateEval,
        SubCriticalTimeVaryingOutFlow, BedFrictionSourceEval,
        RemoveOutofDomainParticles
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator

# PySPH Interpolator
from pysph.tools.interpolator import Interpolator


# Constants
rho_w = 1000.0
g = 9.81
dim = 2


class OkushiriTsunami(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.2,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dx", action="store", type=float, dest="dx", default=0.01875,
            help="Spacing between the particles")
        group.add_argument(
            "--dw0", action="store", type=float, dest="dw0", default=0.13535,
            help="Initial depth of the fluid column (m)")
        group.add_argument(
            "--le", action="store", type=float, dest="le", default=5.448,
            help="Length of the fluid domain (m)")
        group.add_argument(
            "--w", action="store", type=float, dest="w", default=3.402,
            help="Width of the fluid domain (m)")
        group.add_argument(
            "--n", action="store", type=float, dest="n", default=0.025,
            help="Manning coefficient ")
        group.add_argument(
            "--Vb", action="store", type=float, dest="Vb", default=1.96e-4,
            help="Volume of the bed particles")
        group.add_argument(
            "--hb", action="store", type=float, dest="hb", default=1.68e-2,
            help="Smoothing length of the bed particles")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.dw0 = self.options.dw0
        self.le = self.options.le
        self.w = self.options.w
        self.n = self.options.n
        self.Vb = self.options.Vb
        self.hb = self.options.hb

        # Inlet
        self.num_inlet_pa = 2
        self.x_max_inlet = 0
        self.x_min_inlet = -0.95 * self.dx * self.num_inlet_pa
        # Minimum distance allowed between open boundary particles
        self.min_dist_ob = self.dx / 2.

    def initialize(self):
        # Directory containing input files
        self.dir_input_files = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                'okushiri_tsunami_input_files')
        # File containing open boundary conditions
        obc_fname = os.path.join(self.dir_input_files, 'tsunami_obc.txt')
        # Loads the time varying depth field at inlet
        self.t_ob, self.dw_ob = loadtxt(obc_fname, delimiter=' ', unpack=True)

    def create_particles(self):
        hdx = self.hdx
        dx = self.dx
        d = self.dw0
        w = self.w
        y_max = w
        l_domain = self.le

        # Inlet Properties
        y = arange(dx/2, w-(dx/4.), dx)
        x = zeros_like(y) - 0.5*dx

        m = ones_like(x) * dx * dx * rho_w * d
        h = ones_like(x) * hdx * dx
        h0 = ones_like(x) * hdx * dx

        # Stores the time-varying depth field imposed at inlet
        dw_at_t = ones_like(x) * d

        rho = ones_like(x) * rho_w * d
        rho0 = ones_like(x) * rho_w * d
        alpha = dim * rho
        cs = sqrt(9.8 * rho/rho_w)

        inlet = gpa_swe(x=x, y=y, m=m, rho0=rho0, rho=rho, h0=h0, h=h,
                        dw_at_t=dw_at_t, alpha=alpha, cs=cs, name='inlet')
        boundary_props = ['dw_inner_reimann', 'u_inner_reimann',
                          'v_inner_reimann', 'shep_corr']
        inlet.add_output_arrays(boundary_props)

        # Bed Properties
        bed_fname = os.path.join(self.dir_input_files, 'tsunami_bed.txt.bz2')
        xb, yb, b = loadtxt(bed_fname, delimiter=' ', unpack=True)

        Vb = ones_like(xb) * self.Vb
        nb = ones_like(xb) * self.n  # Manning Coefficient
        hb = ones_like(xb) * self.hb

        bed = gpa_swe(name='bed', x=xb, y=yb, V=Vb, n=nb, h=hb, b=b)
        len_b = len(bed.x) * 9
        bed.add_constant('m_mat', [0.0] * len_b)

        # Fluid Properties
        xf, yf = mgrid[0.5*dx:self.x_max_inlet+l_domain:dx,
                       dx/2:y_max-(dx/4.):dx]
        xf, yf = (ravel(t) for t in (xf, yf))
        h = ones_like(xf) * hdx * dx
        h0 = ones_like(xf) * hdx * dx
        fluid = gpa_swe(name='fluid', x=xf, y=yf, h=h, h0=h0)

        compute_fluid_elevation([fluid, bed])

        fluid_surf_hei = d
        dw = fluid_surf_hei - fluid.b
        rho = dw * rho_w
        rho0 = dw * rho_w
        m = rho * dx**2

        fluid.m = m
        fluid.rho = rho
        fluid.rho0 = rho0
        fluid.dw = dw

        # Removes fluid particles with depth less than d_min
        d_min = 7e-5
        idx = where(fluid.dw < d_min)[0]
        fluid.remove_particles(idx)

        # Closed Boundary
        # Note: 5 layers of wall boundary particles
        xcb_top = arange(self.x_min_inlet-5*dx, l_domain+5*dx, dx/2.)
        ycb_top = zeros(0)
        for i in arange(-0.5, 2, 0.5):
            ycb_top = concatenate((ycb_top, ones_like(xcb_top)*(y_max+i*dx)),
                                  axis=0)

        xcb_top = concatenate((xcb_top, xcb_top+dx/4., xcb_top,
                               xcb_top+dx/4., xcb_top), axis=0)

        xcb_bottom = arange(self.x_min_inlet-5*dx, l_domain+5*dx, dx/2.)
        ycb_bottom = zeros(0)
        for i in arange(0, -2.5, -0.5):
            ycb_bottom = concatenate((ycb_bottom, zeros_like(xcb_bottom)+i*dx),
                                     axis=0)

        xcb_bottom = concatenate((xcb_bottom, xcb_bottom+dx/4., xcb_bottom,
                                  xcb_bottom+dx/4., xcb_bottom), axis=0)

        ycb_right = arange(dx/4., y_max-(dx/4.), dx/2.)
        xcb_right = zeros(0)
        for i in arange(0.5, 3.0, 0.5):
            xcb_right = concatenate((xcb_right, zeros_like(ycb_right)
                                     + (l_domain+i*dx)), axis=0)

        ycb_right = concatenate((ycb_right, ycb_right+dx/4., ycb_right,
                                 ycb_right+dx/4, ycb_right), axis=0)

        xcb_all = concatenate((xcb_top, xcb_bottom, xcb_right), axis=0)
        ycb_all = concatenate((ycb_top, ycb_bottom, ycb_right), axis=0)

        m_cb = ones_like(xcb_all) * dx/2. * dx/2. * rho_w * d
        h_cb = ones_like(xcb_all) * hdx * dx/2.
        rho_cb = ones_like(xcb_all) * rho_w * d
        dw_cb = ones_like(xcb_all) * d
        cs_cb = sqrt(9.8 * dw_cb)
        alpha_cb = dim * rho_cb
        is_wall_boun_pa = ones_like(xcb_all)

        boundary = gpa_swe(name='boundary', x=xcb_all, y=ycb_all, m=m_cb,
                           h=h_cb, rho=rho_cb, dw=dw_cb, cs=cs_cb,
                           is_wall_boun_pa=is_wall_boun_pa, alpha=alpha_cb)

        return [inlet, fluid, bed, boundary]

    def create_inlet_outlet(self, particle_arrays):
        f_pa = particle_arrays['fluid']
        i_pa = particle_arrays['inlet']
        b_pa = particle_arrays['bed']
        cb_pa = particle_arrays['boundary']

        inlet = SWEInlet(
            i_pa, f_pa, f_pa, spacing=self.dx, n=self.num_inlet_pa, axis='x',
            xmin=self.x_min_inlet, xmax=self.x_max_inlet, ymin=0, ymax=self.w
        )

        compute_initial_props([i_pa, f_pa, b_pa, cb_pa])

        return [inlet]

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = SWEIntegrator(inlet=SWEInletOutletStep(), fluid=SWEStep())
        tf = 22.51
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            cfl=0.4,
            adaptive_timestep=True,
            output_at_times=(10, 12, 14, 15, 16, 17, 18, 20),
            tf=tf
            )
        return solver

    def pre_step(self, solver):
        t = solver.t
        # Sets the time varying depth for inlet particles based on the current
        # time t
        for pa in self.particles:
            if pa.name == 'inlet':
                for i in range(len(self.t_ob)-1):
                    if t < self.t_ob[i+1]:
                        # Calculates the slope of depth vs time
                        m1 = ((self.dw_ob[i+1]-self.dw_ob[i])
                              / (self.t_ob[i+1]-self.t_ob[i]))
                        # Interpolates the depth based on current time and sets
                        # it for the inlet particles
                        pa.dw_at_t = (ones_like(pa.x)
                                      * (self.dw_ob[i]+m1*(t-self.t_ob[i])))
                        break

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    RemoveFluidParticlesWithNoNeighbors(
                        dest='fluid', sources=['fluid'])
                    ], update_nnps=True
                ),
            Group(
                equations=[
                    RemoveOutofDomainParticles(
                        dest='fluid', x_min=self.x_max_inlet, x_max=self.le,
                        y_min=0, y_max=self.w
                        )
                    ], update_nnps=True
                ),
            Group(
                equations=[
                    RemoveCloseParticlesAtOpenBoundary(
                       min_dist_ob=self.min_dist_ob, dest='inlet',
                       sources=['inlet']
                       )
                    ], update_nnps=True
                ),
            Group(
                equations=[
                    Group(
                        equations=[
                            GatherDensityEvalNextIteration(
                                dest='fluid',
                                sources=['inlet', 'fluid', 'boundary']
                                ),
                            ]
                        ),
                    Group(
                        equations=[
                            NonDimensionalDensityResidual(dest='fluid')
                            ]
                        ),
                    Group(
                        equations=[
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                            ], update_nnps=True
                        ),
                    Group(
                        equations=[
                            CheckConvergenceDensityResidual(dest='fluid')
                            ],
                    )], iterate=True, max_iterations=10
            ),
            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                        dest='fluid', sources=['fluid', 'inlet', 'boundary']),
                        ],
                ),
            Group(
                equations=[
                   RemoveParticlesWithZeroAlpha(dest='fluid'),
                        ], update_nnps=True
                ),
            Group(
                equations=[
                    SWEOS(dest='fluid'),
                    ]
                ),
            Group(
                equations=[
                    BoundaryInnerReimannStateEval(dest='inlet',
                                                  sources=['fluid']),
                    ]
                ),
            Group(
                equations=[
                    SubCriticalTimeVaryingOutFlow(dest='inlet'),
                    ]
                ),
            Group(
                equations=[
                    BedFrictionSourceEval(
                        dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    FluidBottomElevation(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    FluidBottomGradient(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    FluidBottomCurvature(dest='fluid', sources=['bed'])
                    ]
                ),
            Group(
                equations=[
                    ParticleAcceleration(
                        dim=dim, dest='fluid',
                        sources=['fluid', 'inlet', 'boundary'],
                        ),
                    ]
                ),
            ]
        return equations

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')
        self._plot_depth()

    def _plot_depth(self):
        from matplotlib import pyplot as plt

        x_loc_to_interpolate = [4.521, 4.521, 4.521]
        y_loc_to_interpolate = [1.196, 1.696, 2.196]

        # Vacondio simulation results for this case
        fname_vacondio_results = ['tsu_sensor1_vacondio.csv',
                                  'tsu_sensor2_vacondio.csv',
                                  'tsu_sensor3_vacondio.csv']

        output_compare_files_dir = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                'files_for_output_comparison'
                )
        fname_exp_results = os.path.join(output_compare_files_dir,
                                         'tsu_experimental.csv')
        # Experimental values of relative depth at 3 sensor locations
        # corresponding to the coordinates (4.521 m, 1.196 m), (4.521 m, 1.696
        # m), and (4.521 m, 2.196 m)
        # Note: The unit of relative depth in the file is in cm
        t_exp, dw_rel_sens1_exp, dw_rel_sens2_exp, dw_rel_sens3_exp = loadtxt(
                fname_exp_results, delimiter=',', unpack=True)
        dw_relative_sensors_exp = [dw_rel_sens1_exp, dw_rel_sens2_exp,
                                   dw_rel_sens3_exp]

        kernel = CubicSpline(dim=2)
        for n, (x_loc, y_loc) in enumerate(zip(x_loc_to_interpolate,
                                               y_loc_to_interpolate)):
            t_arr = []
            dw_arr = []
            for fname in self.output_files:
                data = load(fname)
                fluid = data['arrays']['fluid']
                t_arr.append(data['solver_data']['t'])
                interp = Interpolator([fluid], kernel=kernel)
                interp.set_interpolation_points(x_loc,
                                                y_loc)
                dw_interp = interp.interpolate('dw')
                dw_arr.append(dw_interp)

            dw_arr = array(dw_arr)
            t_arr = array(t_arr)

            plt.clf()
            if (
                self.dw0 == 0.13535 and
                self.le == 5.448 and
                self.w == 3.402 and
                self.n == 0.025 and
                self.Vb == 1.96e-4
            ):
                fname_vacondio = os.path.join(output_compare_files_dir,
                                              fname_vacondio_results[n])
                t_vac, dw_relative_vac = loadtxt(fname_vacondio,
                                                 delimiter=',',
                                                 unpack=True)

                # Converting the unit of relative depth from cm to m
                dw_relative_exp = dw_relative_sensors_exp[n] / 100.0

                plt.plot(t_vac, dw_relative_vac, 'g.', label='Vacondio et. al')
                plt.plot(t_exp, dw_relative_exp, 'k-',
                         label='Experimental data')
                fname_res = os.path.join(
                    self.output_dir, 'results_sensor%d.npz' % (n+1))
                savez(
                    fname_res,
                    x_sensor=x_loc_to_interpolate[n],
                    y_sensor=y_loc_to_interpolate[n],
                    t_numerical=t_arr, dw_relative_numerical=dw_arr-dw_arr[0],
                    t_vacondio=t_vac, dw_relative_vacondio=dw_relative_vac,
                    t_experimental=t_exp,
                    dw_relative_experimental=dw_relative_exp
                    )

            plt.plot(t_arr, dw_arr-dw_arr[0], 'r.', label='SWESPH')
            plt.legend()
            plt.ylim(-0.015, 0.05)
            plt.xlim(0, max(t_arr))
            plt.xlabel('t (s)')
            plt.ylabel('$dw-dw_0$')
            plt.title(
                    'Relative water depth at the location (%.3f m, %.3f m)' % (
                     x_loc, y_loc)
                    )
            fig = os.path.join(
                    self.output_dir,
                    "relative_depth_sensor"+str(n+1))
            plt.savefig(fig, dpi=300)


def compute_fluid_elevation(particles):
    one_time_equations = [
       Group(
            equations=[
                FluidBottomElevation(dest='fluid', sources=['bed'])
                ]
            ),
       Group(
            equations=[
                GradientCorrectionPreStep(dest='bed', sources=['bed']),
                ]
            ),
       Group(
            equations=[
                GradientCorrection(dest='bed', sources=['bed']),
                ]
            ),
       Group(
            equations=[
                BedGradient(dest='bed', sources=['bed']),
                ]
            ),
       Group(
            equations=[
                BedCurvature(dest='bed', sources=['bed']),
                ]
            ),
    ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


def compute_initial_props(particles):
    one_time_equations = [
                Group(
                    equations=[
                        SWEOS(dest='fluid'),
                        ]
                    ),
            ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == "__main__":
    app = OkushiriTsunami()
    app.run()
    app.post_process(app.info_filename)



"""Definition of some SPH kernel functions
"""

from math import pi, sqrt, exp

M_1_PI = 1.0 / pi
M_2_SQRTPI = 2.0 / sqrt(pi)


def get_correction(kernel, h0):
    rij = kernel.deltap * h0
    return kernel.kernel(rij=rij, h=h0)


def get_compiled_kernel(kernel):
    """Given a kernel, return a high performance wrapper kernel.
    """
    from pysph.base import c_kernels
    cls = getattr(c_kernels, kernel.__class__.__name__)
    wrapper = getattr(c_kernels, kernel.__class__.__name__ + 'Wrapper')
    kern = cls(**kernel.__dict__)
    return wrapper(kern)

###############################################################################
# `CubicSpline` class.
###############################################################################


class CubicSpline(object):
    r"""Cubic Spline Kernel: [Monaghan1992]_

    .. math::
             W(q) = \ &\sigma_3\left[ 1 - \frac{3}{2}q^2\left( 1 -
                    \frac{q}{2} \right) \right],
                    \ & \textrm{for} \ 0 \leq q \leq 1,\\
                  = \ &\frac{\sigma_3}{4}(2-q)^3, & \textrm{for}
                    \ 1 < q \leq 2,\\
                  = \ &0, & \textrm{for}\ q>2, \\

    where :math:`\sigma_3` is a dimensional normalizing factor for the
    cubic spline function given by:

    .. math::
             \sigma_3  = \ & \frac{2}{3h^1}, & \textrm{for dim=1}, \\
             \sigma_3  = \ & \frac{10}{7\pi h^2}, \ & \textrm{for dim=2}, \\
             \sigma_3  = \ & \frac{1}{\pi h^3}, & \textrm{for dim=3}. \\

    References
    ----------
    .. [Monaghan1992] `J. Monaghan, Smoothed Particle Hydrodynamics, "Annual
        Review of Astronomy and Astrophysics", 30 (1992), pp. 543-574.
        <http://adsabs.harvard.edu/abs/1992ARA&A..30..543M>`_
    """

    def __init__(self, dim=1):
        self.radius_scale = 2.0
        self.dim = dim

        if dim == 3:
            self.fac = M_1_PI
        elif dim == 2:
            self.fac = 10 * M_1_PI / 7.0
        else:
            self.fac = 2.0 / 3.0

    def get_deltap(self):
        return 2. / 3

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        tmp2 = 2. - q
        if (q > 2.0):
            val = 0.0

        elif (q > 1.0):
            val = 0.25 * tmp2 * tmp2 * tmp2
        else:
            val = 1 - 1.5 * q * q * (1 - 0.5 * q)

        return val * fac

    def dwdq(self, rij=1.0, h=1.0):
        """Gradient of a kernel is given by
        .. math::
            \nabla W = normalization  \frac{dW}{dq} \frac{dq}{dx}
            \nabla W = w_dash  \frac{dq}{dx}

        Here we get `w_dash` by using `dwdq` method
        """
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor ( sigma )
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute sigma * dw_dq
        tmp2 = 2. - q
        if (rij > 1e-12):
            if (q > 2.0):
                val = 0.0
            elif (q > 1.0):
                val = -0.75 * tmp2 * tmp2
            else:
                val = -3.0 * q * (1 - 0.75 * q)
        else:
            val = 0.0

        return val * fac

    def gradient(self, xij=[0., 0, 0], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h
        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # kernel and gradient evaluated at q
        tmp2 = 2. - q
        if (q > 2.0):
            w = 0.0
            dw = 0.0

        elif (q > 1.0):
            w = 0.25 * tmp2 * tmp2 * tmp2
            dw = -0.75 * tmp2 * tmp2
        else:
            w = 1 - 1.5 * q * q * (1 - 0.5 * q)
            dw = -3.0 * q * (1 - 0.75 * q)

        return -fac * h1 * (dw * q + w * self.dim)


class WendlandQuinticC2_1D(object):
    r"""The following is the WendlandQuintic kernel (Wendland C2) kernel for 1D.

    .. math::
             W(q) = \ & \alpha_d (1-q/2)^3 (1.5q +1))),
                    \ & \textrm{for} \ 0\leq q \leq 2,\\
                  = \ & 0, & \textrm{for} \ q>2,\\

    where :math:`d` is the number of dimensions and

    .. math::
             \alpha_d  =  \frac{5}{8h}, \textrm{for dim=1}


    """
    def __init__(self, dim=1):
        self.radius_scale = 2.0
        self.dim = dim
        if dim == 1:
            self.fac = 5.0 / 8.0
        elif dim == 2:
            raise ValueError(
                "WendlandQuinticC2_1D: Dim %d not supported" % dim)
        elif dim == 3:
            raise ValueError(
                "WendlandQuinticC2_1D: Dim %d not supported" % dim)

    def get_deltap(self):
        return 2.0/3

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * (1.5 * q + 1.0)

        return val * fac

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -3.0 * q * tmp * tmp

        return val * fac

    def gradient(self, xij=[0., 0, 0], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * (1.5 * q + 1.0)
            dw = -3.0 * q * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)


class WendlandQuintic(object):
    r"""The following is the WendlandQuintic kernel(C2) kernel for 2D and 3D.

    .. math::
             W(q) = \ & \alpha_d (1-q/2)^4(2q +1))),
                    \ & \textrm{for} \ 0\leq q \leq 2,\\
                  = \ & 0, & \textrm{for} \ q>2,\\

    where :math:`d` is the number of dimensions and

    .. math::
             \alpha_d  = \ & \frac{7}{4\pi h^2}, \ & \textrm{for dim=2}, \\
             \alpha_d  = \ & \frac{21}{16\pi h^3}, \ & \textrm{for dim=3}

    """

    def __init__(self, dim=2):
        self.radius_scale = 2.0
        if dim == 1:
            raise ValueError("WendlandQuintic: Dim %d not supported" % dim)
        self.dim = dim

        if dim == 2:
            self.fac = 7.0 * M_1_PI / 4.0
        elif dim == 3:
            self.fac = M_1_PI * 21.0 / 16.0

    def get_deltap(self):
        return 0.5

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * (2.0 * q + 1.0)

        return val * fac

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -5.0 * q * tmp * tmp * tmp

        return val * fac

    def gradient(self, xij=[0., 0, 0], rij=1.0, h=1.0,
                 grad=[0, 0, 0]):
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * (2.0 * q + 1.0)
            dw = -5.0 * q * tmp * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)


class WendlandQuinticC4_1D(object):
    r"""The following is the WendlandQuintic kernel (Wendland C4) kernel for 1D.

    .. math::
             W(q) = \ & \alpha_d (1-q/2)^5 (2q^2 + 2.5q +1))),
                     \ & \textrm{for} \ 0\leq q \leq 2,\\
                  = \ & 0, & \textrm{for} \ q>2,\\

    where :math:`d` is the number of dimensions and

    .. math::
            \alpha_d =  \frac{3}{4h}, \ \textrm{for dim=1}

    """

    def __init__(self, dim=1):
        self.radius_scale = 2.0
        self.dim = dim
        if dim == 1:
            self.fac = 0.75

        if dim == 2:
            raise ValueError(
                "WendlandQuinticC4_1D: Dim %d not supported" % dim)

        elif dim == 3:
            raise ValueError(
                "WendlandQuinticC4_1D: Dim %d not supported" % dim)

    def get_deltap(self):
        return 0.55195628

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * tmp * (2 * q * q + 2.5 * q + 1.0)

        return val * fac

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -3.5 * q * (2 * q + 1) * tmp * tmp * tmp * tmp

        return val * fac

    def gradient(self, xij=[0., 0., 0.], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * tmp * (2 * q * q + 2.5 * q + 1.0)
            dw = -3.5 * q * (2 * q + 1) * tmp * tmp * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)


class WendlandQuinticC4(object):
    r"""The following is the WendlandQuintic kernel (Wendland C4) kernel for
        2D and 3D.

    .. math::
             W(q) = \ & \alpha_d (1-q/2)^6(\frac{35}{12} q^2 + 3q +1))),
                     \ & \textrm{for} \ 0\leq q \leq 2,\\
                  = \ & 0, & \textrm{for} \ q>2,\\

    where :math:`d` is the number of dimensions and

    .. math::
             \alpha_d  = \ & \frac{9}{4\pi h^2}, \ & \textrm{for dim=2}, \\
             \alpha_d  = \ & \frac{495}{256\pi h^3}, \ & \textrm{for dim=3}

    """

    def __init__(self, dim=2):
        self.radius_scale = 2.0
        self.dim = dim
        if dim == 1:
            raise ValueError("WendlandQuinticC4: Dim %d not supported" % dim)

        if dim == 2:
            self.fac = 9.0 * M_1_PI / 4.0
        elif dim == 3:
            self.fac = M_1_PI * 495.0 / 256.0

    def get_deltap(self):
        return 0.47114274

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * tmp * tmp * \
                ((35.0 / 12.0) * q * q + 3.0 * q + 1.0)

        return val * fac

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = (-14.0 / 3.0) * q * (1 + 2.5 * q) * \
                    tmp * tmp * tmp * tmp * tmp

        return val * fac

    def gradient(self, xij=[0., 0., 0.], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * tmp * tmp * \
                ((35.0 / 12.0) * q * q + 3.0 * q + 1.0)
            dw = (-14.0 / 3.0) * q * (1 + 2.5 * q) * \
                tmp * tmp * tmp * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)


class WendlandQuinticC6_1D(object):
    r"""The following is the WendlandQuintic kernel (Wendland C6) kernel for 1D.

    .. math::
             W(q) = \ & \alpha_d (1-q/2)^7 (\frac{21}{8} q^3
                    + \frac{19}{4} q^2 + 3.5q +1))),
                    \ & \textrm{for} \ 0\leq q \leq 2,\\
                  = \ & 0, & \textrm{for} \ q>2,\\

    where :math:`d` is the number of dimensions and

    .. math::
             \alpha_d  = \ \frac{55}{64h},  \textrm{for dim=1}

    """

    def __init__(self, dim=1):
        self.radius_scale = 2.0
        self.dim = dim
        if dim == 1:
            self.fac = 55.0 / 64.0

        if dim == 2:
            raise ValueError(
                "WendlandQuinticC6_1D: Dim %d not supported" % dim)
        elif dim == 3:
            raise ValueError(
                "WendlandQuinticC6_1D: Dim %d not supported" % dim)

    def get_deltap(self):
        return 0.47996698

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * tmp * tmp * tmp * \
                (2.625 * q * q * q + 4.75 * q * q + 3.5 * q + 1.0)

        return val * fac

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -0.5 * q * (26.25 * q * q + 27 * q + 9.0) * \
                    tmp * tmp * tmp * tmp * tmp * tmp

        return val * fac

    def gradient(self, xij=[0., 0., 0.], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h
        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * tmp * tmp * tmp * \
                (2.625 * q * q * q + 4.75 * q * q + 3.5 * q + 1.0)
            dw = -0.5 * q * (26.25 * q * q + 27 * q + 9.0) * \
                tmp * tmp * tmp * tmp * tmp * tmp

        return -fac * h1 * (dw * q + w * self.dim)


class WendlandQuinticC6(object):
    r"""The following is the WendlandQuintic kernel(C6) kernel for 2D and 3D.

    .. math::
             W(q) = \ & \alpha_d (1-q/2)^8 (4 q^3 + 6.25 q^2 + 4q +1))),
                     \ & \textrm{for} \ 0\leq q \leq 2,\\
                  = \ & 0, & \textrm{for} \ q>2,\\

    where :math:`d` is the number of dimensions and

    .. math::
             \alpha_d  = \ & \frac{78}{28\pi h^2}, \ & \textrm{for dim=2}, \\
             \alpha_d  = \ & \frac{1365}{512\pi h^3}, \ & \textrm{for dim=3}

    """

    def __init__(self, dim=2):
        self.radius_scale = 2.0
        self.dim = dim
        if dim == 1:
            raise ValueError("WendlandQuinticC6: Dim %d not supported" % dim)

        if dim == 2:
            self.fac = 78.0 * M_1_PI / 28.0
        elif dim == 3:
            self.fac = M_1_PI * 1365.0 / 512.0

    def get_deltap(self):
        return 0.4305720757

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1.0 / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0

        tmp = 1. - 0.5 * q
        if (q < 2.0):
            val = tmp * tmp * tmp * tmp * tmp * tmp * tmp * tmp * \
                (4.0 * q * q * q + 6.25 * q * q + 4.0 * q + 1.0)

        return val * fac

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            if (rij > 1e-12):
                val = -5.50 * q * tmp * tmp * tmp * tmp * tmp * \
                    tmp * tmp * (1.0 + 3.5 * q + 4 * q * q)

        return val * fac

    def gradient(self, xij=[0., 0., 0.], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h
        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the kernel and gradient at q
        w = 0.0
        dw = 0.0
        tmp = 1.0 - 0.5 * q
        if (q < 2.0):
            w = tmp * tmp * tmp * tmp * tmp * tmp * tmp * tmp * \
                (4.0 * q * q * q + 6.25 * q * q + 4.0 * q + 1.0)
            dw = -5.50 * q * tmp * tmp * tmp * tmp * tmp * \
                tmp * tmp * (1.0 + 3.5 * q + 4 * q * q)

        return -fac * h1 * (dw * q + w * self.dim)


class Gaussian(object):
    r"""Gaussian Kernel: [Liu2010]_

    .. math::
             W(q) = \ &\sigma_g e^{-q^2}, \ & \textrm{for} \ 0\leq q \leq 3,\\
                  = \ & 0, & \textrm{for} \ q>3,\\

    where :math:`\sigma_g` is a dimensional normalizing factor for the gaussian
    function given by:

    .. math::
             \sigma_g  = \ & \frac{1}{\pi^{1/2} h}, \ & \textrm{for dim=1}, \\
             \sigma_g  = \ & \frac{1}{\pi h^2}, \ & \textrm{for dim=2}, \\
             \sigma_g  = \ & \frac{1}{\pi^{3/2} h^3}, & \textrm{for dim=3}. \\

    References
    ----------
    .. [Liu2010] `M. Liu, & G. Liu, Smoothed particle hydrodynamics (SPH):
        an overview and recent developments, "Archives of computational
        methods in engineering", 17.1 (2010), pp. 25-76.
        <http://link.springer.com/article/10.1007/s11831-010-9040-7>`_
    """

    def __init__(self, dim=2):
        self.radius_scale = 3.0
        self.dim = dim

        self.fac = 0.5 * M_2_SQRTPI
        if dim > 1:
            self.fac *= 0.5 * M_2_SQRTPI
        if dim > 2:
            self.fac *= 0.5 * M_2_SQRTPI

    def get_deltap(self):
        # The inflection point is at q=1/sqrt(2)
        # the deltap values for some standard kernels
        # have been tabulated in sec 3.2 of
        # http://cfd.mace.manchester.ac.uk/sph/SPH_PhDs/2008/crespo_thesis.pdf
        return 0.70710678118654746

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0
        if (q < 3.0):
            val = exp(-q * q) * fac

        return val

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        if (q < 3.0):
            if (rij > 1e-12):
                val = -2.0 * q * exp(-q * q)

        return val * fac

    def gradient(self, xij=[0., 0., 0.], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0., 0.], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # kernel and gradient evaluated at q
        w = 0.0
        dw = 0.0
        if (q < 3.0):
            w = exp(-q * q)
            dw = -2.0 * q * w

        return -fac * h1 * (dw * q + w * self.dim)


class SuperGaussian(object):
    r"""Super Gaussian Kernel: [Monaghan1992]_

    .. math::
             W(q) = \ &\frac{1}{h^{d}\pi^{d/2}} e^{-q^2} (d/2 + 1 - q^2),
                    \ & \textrm{for} \ 0\leq q \leq 3,\\
                  = \ & 0, & \textrm{for} \ q>3,\\

    where :math:`d` is the number of dimensions.
    """

    def __init__(self, dim=2):
        self.radius_scale = 3.0
        self.dim = dim

        self.fac = 0.5 * M_2_SQRTPI
        if dim > 1:
            self.fac *= 0.5 * M_2_SQRTPI
        if dim > 2:
            self.fac *= 0.5 * M_2_SQRTPI

    def get_deltap(self):
        # Found inflection point using sympy.
        if self.dim == 1:
            return 0.584540507426389
        elif self.dim == 2:
            return 0.6021141014644256
        else:
            return 0.615369528365158

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        val = 0.0
        if (q < 3.0):
            q2 = q * q
            val = exp(-q2) * (1.0 + self.dim * 0.5 - q2) * fac

        return val

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        # compute the gradient
        val = 0.0
        if (q < 3.0):
            if (rij > 1e-12):
                q2 = q * q
                val = q * (2.0 * q2 - self.dim - 4) * exp(-q2)

        return val * fac

    def gradient(self, xij=[0., 0., 0.], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h
        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0., 0.], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1
        d = self.dim

        # get the kernel normalizing factor
        if d == 1:
            fac = self.fac * h1
        elif d == 2:
            fac = self.fac * h1 * h1
        elif d == 3:
            fac = self.fac * h1 * h1 * h1

        # kernel and gradient evaluated at q
        val = 0.0
        if (q < 3.0):
            q2 = q * q
            val = (-d * d * 0.5 + 2.0 * d * q2 - d -
                   2.0 * q2 * q2 + 4 * q2) * exp(-q2)

        return -fac * h1 * val


class QuinticSpline(object):
    r"""Quintic Spline SPH kernel: [Liu2010]_

    .. math::
             W(q) = \ &\sigma_5\left[ (3-q)^5 - 6(2-q)^5 + 15(1-q)^5 \right],
                    \ & \textrm{for} \ 0\leq q \leq 1,\\
                  = \ &\sigma_5\left[ (3-q)^5 - 6(2-q)^5 \right], &
                    \textrm{for} \ 1 <  q \leq 2,\\
                  = \ &\sigma_5 \ (3-q)^5 , & \textrm{for} \ 2 < q \leq 3,\\
                  = \ & 0, & \textrm{for} \ q>3,\\

    where :math:`\sigma_5` is a dimensional normalizing factor for the
    quintic spline function given by:

    .. math::
             \sigma_5  = \ & \frac{1}{120 h^1}, & \textrm{for dim=1}, \\
             \sigma_5  = \ & \frac{7}{478\pi h^2}, \ & \textrm{for dim=2}, \\
             \sigma_5  = \ & \frac{1}{120\pi h^3}, & \textrm{for dim=3}. \\

    """

    def __init__(self, dim=2):
        self.radius_scale = 3.0
        self.dim = dim

        if dim == 1:
            self.fac = 1.0 / 120.0
        elif dim == 2:
            self.fac = M_1_PI * 7.0 / 478.0
        elif dim == 3:
            self.fac = M_1_PI * 1.0 / 120.0

    def get_deltap(self):
        # The inflection points for the polynomial are obtained as
        # http://www.wolframalpha.com/input/?i=%28%283-x%29%5E5+-+6*%282-x%29%5E5+%2B+15*%281-x%29%5E5%29%27%27
        # the only permissible value is taken
        return 0.759298480738450

    def kernel(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        tmp3 = 3. - q
        tmp2 = 2. - q
        tmp1 = 1. - q
        if (q > 3.0):
            val = 0.0

        elif (q > 2.0):
            val = tmp3 * tmp3 * tmp3 * tmp3 * tmp3

        elif (q > 1.0):
            val = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            val -= 6.0 * tmp2 * tmp2 * tmp2 * tmp2 * tmp2

        else:
            val = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            val -= 6.0 * tmp2 * tmp2 * tmp2 * tmp2 * tmp2
            val += 15. * tmp1 * tmp1 * tmp1 * tmp1 * tmp1

        return val * fac

    def dwdq(self, rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        tmp3 = 3. - q
        tmp2 = 2. - q
        tmp1 = 1. - q

        # compute the gradient
        if (rij > 1e-12):
            if (q > 3.0):
                val = 0.0

            elif (q > 2.0):
                val = -5.0 * tmp3 * tmp3 * tmp3 * tmp3

            elif (q > 1.0):
                val = -5.0 * tmp3 * tmp3 * tmp3 * tmp3
                val += 30.0 * tmp2 * tmp2 * tmp2 * tmp2
            else:
                val = -5.0 * tmp3 * tmp3 * tmp3 * tmp3
                val += 30.0 * tmp2 * tmp2 * tmp2 * tmp2
                val -= 75.0 * tmp1 * tmp1 * tmp1 * tmp1
        else:
            val = 0.0

        return val * fac

    def gradient(self, xij=[0., 0., 0.], rij=1.0, h=1.0, grad=[0, 0, 0]):
        h1 = 1. / h

        # compute the gradient.
        if (rij > 1e-12):
            wdash = self.dwdq(rij, h)
            tmp = wdash * h1 / rij
        else:
            tmp = 0.0

        grad[0] = tmp * xij[0]
        grad[1] = tmp * xij[1]
        grad[2] = tmp * xij[2]

    def gradient_h(self, xij=[0., 0, 0], rij=1.0, h=1.0):
        h1 = 1. / h
        q = rij * h1

        # get the kernel normalizing factor
        if self.dim == 1:
            fac = self.fac * h1
        elif self.dim == 2:
            fac = self.fac * h1 * h1
        elif self.dim == 3:
            fac = self.fac * h1 * h1 * h1

        tmp3 = 3. - q
        tmp2 = 2. - q
        tmp1 = 1. - q

        # compute the kernel & gradient at q
        if (q > 3.0):
            w = 0.0
            dw = 0.0

        elif (q > 2.0):
            w = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            dw = -5.0 * tmp3 * tmp3 * tmp3 * tmp3

        elif (q > 1.0):
            w = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            w -= 6.0 * tmp2 * tmp2 * tmp2 * tmp2 * tmp2

            dw = -5.0 * tmp3 * tmp3 * tmp3 * tmp3
            dw += 30.0 * tmp2 * tmp2 * tmp2 * tmp2
        else:
            w = tmp3 * tmp3 * tmp3 * tmp3 * tmp3
            w -= 6.0 * tmp2 * tmp2 * tmp2 * tmp2 * tmp2
            w += 15. * tmp1 * tmp1 * tmp1 * tmp1 * tmp1

            dw = -5.0 * tmp3 * tmp3 * tmp3 * tmp3
            dw += 30.0 * tmp2 * tmp2 * tmp2 * tmp2
            dw -= 75.0 * tmp1 * tmp1 * tmp1 * tmp1

        return -fac * h1 * (dw * q + w * self.dim)



"""Demonstrate the inlet and outlet feature in 2D. (1 second)

We first create three particle arrays, an "inlet", "fluid" and "outlet" in the
`create_particle` method. Initially there are no fluid particles. A block
of inlet and outlet particles are created. The inlet is created in the
region (-1.0, 0.0) and (0.0, 1.0). velocity is prescribed to the inlet i.e.
along the y-axis with a u velocity = 0.25.

An outlet is also created in the region (1.0, 2.0), (0.0, 1.0) and as fluid
particles enter the outlet region, they are converted to outlet particles.  As
outlet particles leave the outlet they are removed from the simulation.

The `SimpleInletOutlet` is created which created `Inlet` and `Outlet` objects
which can update the particles when they are moved from inlet to outlet. Also
other missing variables in `InletInfo` and `OutletInfo` are evaluated by the
manager.


The following figure should make this clear.

               inlet       fluid       outlet
              ---------    --------    --------
             | * * * * |  |        |  | * * * *|
     u       | * * * * |  |        |  | * * * *|
    --->     | * * * * |  |        |  | * * * *|
             | * * * * |  |        |  | * * * *|
              --------     --------    --------

In the figure '*' are the particles at the t=0. The particles are moving to the
right and as they do, new fluid particles are added and as the fluid particles
flow into the outlet they are converted to the outlet particle array and at
last as the particles leave the outlet they are removed from the simulation.

This example can be run in parallel.

"""

import numpy as np

from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.sph.integrator import PECIntegrator
from pysph.sph.bc.donothing.simple_inlet_outlet import (
    SimpleInletOutlet)
from pysph.sph.bc.inlet_outlet_manager import (
    InletInfo, OutletInfo, OutletStep, InletStep)
from pysph.sph.basic_equations import SummationDensity


class InletOutletApp(Application):

    def add_user_options(self, group):
        group.add_argument(
            "--speed", action="store", type=float, dest="speed",
            default=0.25, help="Speed of inlet particles.")

    def create_particles(self):
        # Note that you need to create the inlet and outlet arrays
        # in this method.

        # Initially fluid has no particles -- these are generated by the inlet.
        fluid = get_particle_array(name='fluid')

        # Setup the inlet particle array with just the particles we need at the
        dx = 0.1
        x, y = np.mgrid[-1+dx/2: 0: dx, 0:1:dx]
        m = np.ones_like(x)*dx*dx
        h = np.ones_like(x)*dx*1.5
        rho = np.ones_like(x)

        # Remember to set u otherwise the inlet particles won't move.  Here we
        # use the options which may be set by the user from the command line.
        u = np.ones_like(x)*self.options.speed

        inlet = get_particle_array(name='inlet', x=x, y=y, m=m, h=h, u=u,
                                   rho=rho)
        x += 2.0
        outlet = get_particle_array(name='outlet', x=x, y=y, m=m, h=h, u=u,
                                    rho=rho)

        particles = [inlet, fluid, outlet]

        props = ['ioid', 'disp', 'x0']
        for p in props:
            for pa in particles:
                pa.add_property(p)

        return particles

    def _create_inlet_outlet_manager(self):
        from pysph.sph.bc.donothing.inlet import Inlet
        from pysph.sph.bc.donothing.outlet import Outlet

        props_to_copy = ['x', 'y', 'z', 'u', 'v', 'w', 'm',
                         'h', 'rho', 'p', 'ioid']
        inlet_info = InletInfo(
            pa_name='inlet', normal=[-1.0, 0.0, 0.0],
            refpoint=[0.0, 0.0, 0.0], has_ghost=False,
            update_cls=Inlet
        )

        outlet_info = OutletInfo(
            pa_name='outlet', normal=[1.0, 0.0, 0.0],
            refpoint=[1.0, 0.0, 0.0], update_cls=Outlet,
            props_to_copy=props_to_copy
        )

        iom = SimpleInletOutlet(
            fluid_arrays=['fluid'], inletinfo=[inlet_info],
            outletinfo=[outlet_info]
        )

        return iom

    def create_inlet_outlet(self, particle_arrays):
        iom = self.iom
        io = iom.get_inlet_outlet(particle_arrays)
        return io

    def create_equations(self):
        equations = [
            SummationDensity(
                dest='fluid', sources=['inlet', 'outlet', 'fluid']
            )
        ]
        return equations

    def create_solver(self):
        self.iom = self._create_inlet_outlet_manager()
        kernel = CubicSpline(dim=2)
        integrator = PECIntegrator(
            fluid=InletStep(), inlet=InletStep(),
            outlet=OutletStep()
        )
        self.iom.active_stages = [2]
        self.iom.setup_iom(dim=2, kernel=kernel)
        self.iom.update_dx(dx=0.1)
        dt = 1e-2
        tf = 12

        solver = Solver(
            kernel=kernel, dim=2, integrator=integrator, dt=dt, tf=tf,
            adaptive_timestep=False, pfreq=20
        )
        return solver


if __name__ == '__main__':
    app = InletOutletApp()
    app.run()



"""Summation density example using the cell based NNPS partitioner."""

import mpi4py.MPI as mpi

import numpy
from numpy import random

# Carray from PyZoltan
from cyarray.carray import UIntArray

# PySPH imports
from pysph.base.nnps import BoxSortNNPS
from pysph.parallel.parallel_manager import ZoltanParallelManagerGeometric
from pysph.base.utils import get_particle_array_wcsph

from pysph.base.kernels import CubicSpline, get_compiled_kernel
"""Utility to compute summation density"""


dim = 3


def sd_evaluate(nnps, pm, mass, src_index, dst_index):
    # the destination particle array
    dst = nnps.particles[dst_index]
    src = nnps.particles[src_index]

    # particle coordinates
    dx, dy, dz, dh, drho = dst.get(
        'x', 'y', 'z', 'h', 'rho', only_real_particles=True)
    sx, sy, sz, sh, srho = src.get(
        'x', 'y', 'z', 'h', 'rho', only_real_particles=False)

    neighbors = UIntArray()
    cubic = get_compiled_kernel(CubicSpline(dim=dim))

    # compute density for each destination particle
    num_particles = dst.num_real_particles

    # the number of local particles should have tag Local
    assert (num_particles == pm.num_local[dst_index])

    for i in range(num_particles):

        hi = dh[i]

        nnps.get_nearest_particles(src_index, dst_index, i, neighbors)

        nnbrs = neighbors.length

        rho_sum = 0.0
        for indexj in range(nnbrs):
            j = neighbors[indexj]

            wij = cubic.kernel(dx[i], dy[i], dz[i], sx[j], sy[j], sz[j], hi)

            rho_sum = rho_sum + mass * wij

        drho[i] += rho_sum


def main():
    # Initialize MPI and find out number of local particles
    comm = mpi.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    # number of particles per array
    numMyPoints = 1 << 10
    numGlobalPoints = size * numMyPoints

    avg_vol = 1.0 / numGlobalPoints
    dx = numpy.power(avg_vol, 1.0 / dim)
    mass = avg_vol
    hdx = 1.3

    if numGlobalPoints % size != 0:
        raise RuntimeError("Run with 2^n num procs!")

    # everybody creates two particle arrays with numMyPoints
    x1 = random.random(numMyPoints)
    y1 = random.random(numMyPoints)
    z1 = random.random(numMyPoints)
    h1 = numpy.ones_like(x1) * hdx * dx
    rho1 = numpy.zeros_like(x1)

    x2 = random.random(numMyPoints)
    y2 = random.random(numMyPoints)
    z2 = random.random(numMyPoints)
    h2 = numpy.ones_like(x2) * hdx * dx
    rho2 = numpy.zeros_like(x2)

    # z1[:] = 1.0
    # z2[:] = 0.5

    # local particle arrays
    pa1 = get_particle_array_wcsph(x=x1, y=y1, h=h1, rho=rho1, z=z1)
    pa2 = get_particle_array_wcsph(x=x2, y=y2, h=h2, rho=rho2, z=z2)

    # gather the data on root
    X1 = numpy.zeros(numGlobalPoints)
    Y1 = numpy.zeros(numGlobalPoints)
    Z1 = numpy.zeros(numGlobalPoints)
    H1 = numpy.ones_like(X1) * hdx * dx
    RHO1 = numpy.zeros_like(X1)

    gathers = (numpy.ones(size) * numMyPoints, None)

    comm.Gatherv(sendbuf=[x1, mpi.DOUBLE], recvbuf=[X1, gathers, mpi.DOUBLE])
    comm.Gatherv(sendbuf=[y1, mpi.DOUBLE], recvbuf=[Y1, gathers, mpi.DOUBLE])
    comm.Gatherv(sendbuf=[z1, mpi.DOUBLE], recvbuf=[Z1, gathers, mpi.DOUBLE])
    comm.Gatherv(
        sendbuf=[rho1, mpi.DOUBLE], recvbuf=[RHO1, gathers, mpi.DOUBLE]
    )

    X2 = numpy.zeros(numGlobalPoints)
    Y2 = numpy.zeros(numGlobalPoints)
    Z2 = numpy.zeros(numGlobalPoints)
    H2 = numpy.ones_like(X2) * hdx * dx
    RHO2 = numpy.zeros_like(X2)

    comm.Gatherv(sendbuf=[x2, mpi.DOUBLE], recvbuf=[X2, gathers, mpi.DOUBLE])
    comm.Gatherv(sendbuf=[y2, mpi.DOUBLE], recvbuf=[Y2, gathers, mpi.DOUBLE])
    comm.Gatherv(sendbuf=[z2, mpi.DOUBLE], recvbuf=[Z2, gathers, mpi.DOUBLE])
    comm.Gatherv(
        sendbuf=[rho2, mpi.DOUBLE], recvbuf=[RHO2, gathers, mpi.DOUBLE]
    )

    # create the particle arrays and PM
    PA1 = get_particle_array_wcsph(x=X1, y=Y1, z=Z1, h=H1, rho=RHO1)
    PA2 = get_particle_array_wcsph(x=X2, y=Y2, z=Z2, h=H2, rho=RHO2)

    # create the parallel manager
    PARTICLES = [PA1, PA2]
    PM = ZoltanParallelManagerGeometric(
        dim=dim, particles=PARTICLES, comm=comm
    )

    # create the local NNPS object with all the particles
    Nnps = BoxSortNNPS(dim=dim, particles=PARTICLES)
    Nnps.update()

    # only root computes summation density
    if rank == 0:
        assert numpy.allclose(PA1.rho, 0)
        sd_evaluate(Nnps, PM, mass, src_index=1, dst_index=0)
        sd_evaluate(Nnps, PM, mass, src_index=0, dst_index=0)
        RHO1 = PA1.rho

        assert numpy.allclose(PA2.rho, 0)
        sd_evaluate(Nnps, PM, mass, src_index=0, dst_index=1)
        sd_evaluate(Nnps, PM, mass, src_index=1, dst_index=1)
        RHO2 = PA2.rho

    # wait for the root...
    comm.barrier()

    # create the local particle arrays
    particles = [pa1, pa2]

    # create the local nnps object and parallel manager
    pm = ZoltanParallelManagerGeometric(
        dim=dim, comm=comm, particles=particles
    )
    nnps = BoxSortNNPS(dim=dim, particles=particles)

    # set the Zoltan parameters (Optional)
    pz = pm.pz
    pz.set_lb_method("RCB")
    pz.Zoltan_Set_Param("DEBUG_LEVEL", "0")

    # Update the parallel manager (distribute particles)
    pm.update()

    # update the local nnps
    nnps.update()

    # Compute summation density individually on each processor
    sd_evaluate(nnps, pm, mass, src_index=0, dst_index=1)
    sd_evaluate(nnps, pm, mass, src_index=1, dst_index=1)

    sd_evaluate(nnps, pm, mass, src_index=0, dst_index=0)
    sd_evaluate(nnps, pm, mass, src_index=1, dst_index=0)

    # gather the density and global ids
    rho1 = pa1.rho
    tmp = comm.gather(rho1)
    if rank == 0:
        global_rho1 = numpy.concatenate(tmp)
        assert (global_rho1.size == numGlobalPoints)

    rho2 = pa2.rho
    tmp = comm.gather(rho2)
    if rank == 0:
        global_rho2 = numpy.concatenate(tmp)
        assert (global_rho2.size == numGlobalPoints)

    # gather global x1 and y1
    x1 = pa1.x
    tmp = comm.gather(x1)
    if rank == 0:
        global_x1 = numpy.concatenate(tmp)
        assert (global_x1.size == numGlobalPoints)

    y1 = pa1.y
    tmp = comm.gather(y1)
    if rank == 0:
        global_y1 = numpy.concatenate(tmp)
        assert (global_y1.size == numGlobalPoints)

    z1 = pa1.z
    tmp = comm.gather(z1)
    if rank == 0:
        global_z1 = numpy.concatenate(tmp)
        assert (global_z1.size == numGlobalPoints)

    # gather global x2 and y2
    x2 = pa2.x
    tmp = comm.gather(x2)
    if rank == 0:
        global_x2 = numpy.concatenate(tmp)
        assert (global_x2.size == numGlobalPoints)

    y2 = pa2.y
    tmp = comm.gather(y2)
    if rank == 0:
        global_y2 = numpy.concatenate(tmp)
        assert (global_y2.size == numGlobalPoints)

    z2 = pa2.z
    tmp = comm.gather(z2)
    if rank == 0:
        global_z2 = numpy.concatenate(tmp)
        assert (global_z2.size == numGlobalPoints)

    # gather global indices
    gid1 = pa1.gid
    tmp = comm.gather(gid1)
    if rank == 0:
        global_gid1 = numpy.concatenate(tmp)
        assert (global_gid1.size == numGlobalPoints)

    gid2 = pa2.gid
    tmp = comm.gather(gid2)
    if rank == 0:
        global_gid2 = numpy.concatenate(tmp)
        assert (global_gid2.size == numGlobalPoints)

    # check rho1
    if rank == 0:
        # make sure the arrays are of the same size
        assert (global_x1.size == X1.size)
        assert (global_y1.size == Y1.size)
        assert (global_z1.size == Z1.size)

        for i in range(numGlobalPoints):

            # make sure we're chacking the right point
            assert abs(global_x1[i] - X1[global_gid1[i]]) < 1e-14
            assert abs(global_y1[i] - Y1[global_gid1[i]]) < 1e-14
            assert abs(global_z1[i] - Z1[global_gid1[i]]) < 1e-14

            diff = abs(global_rho1[i] - RHO1[global_gid1[i]])
            condition = diff < 1e-14
            assert condition, "diff = %g" % (diff)

    # check rho2
    if rank == 0:
        # make sure the arrays are of the same size
        assert (global_x2.size == X2.size)
        assert (global_y2.size == Y2.size)
        assert (global_z2.size == Z2.size)

        for i in range(numGlobalPoints):

            # make sure we're chacking the right point
            assert abs(global_x2[i] - X2[global_gid2[i]]) < 1e-14
            assert abs(global_y2[i] - Y2[global_gid2[i]]) < 1e-14
            assert abs(global_z2[i] - Z2[global_gid2[i]]) < 1e-14

            diff = abs(global_rho2[i] - RHO2[global_gid2[i]])
            condition = diff < 1e-14
            assert condition, "diff = %g" % (diff)

    if rank == 0:
        print("Summation density test: OK")


if __name__ == '__main__':
    main()



#cython: embedsignature=True

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.vector cimport vector
from libcpp.map cimport map
from libc.stdlib cimport bsearch
from libc.stdio cimport printf

from cython.operator cimport dereference as deref, preincrement as inc

# Cython for compiler directives
cimport cython

import numpy as np
cimport numpy as np

cdef extern from "<algorithm>" namespace "std" nogil:
    void sort[Iter, Compare](Iter first, Iter last, Compare comp)
    void sort[Iter](Iter first, Iter last)

#############################################################################

cdef inline int cmp_func(const void* a, const void* b) nogil:
    return (<uint64_t*>a)[0] - (<uint64_t*>b)[0]

cdef class ZOrderNNPS(NNPS):

    """Find nearest neighbors using Z-Order space filling curve"""

    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int H=1,
            bint asymmetric=False):
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        self.radius_scale2 = radius_scale*radius_scale
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int i, num_particles

        self.asymmetric = asymmetric

        self.H = H
        self.mask_len = (2 * H + 1) ** 3

        self.src_index = 0
        self.dst_index = 0
        self.sort_gids = sort_gids
        self.domain.update()
        self.update()

    def __cinit__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int H=1,
            bint asymmetric=False):
        cdef int narrays = len(particles)

        self.pids = <uint32_t**> malloc(narrays*sizeof(uint32_t*))
        self.keys = <uint64_t**> malloc(narrays*sizeof(uint64_t*))
        self.cids = <uint32_t**> malloc(narrays*sizeof(uint32_t*))
        self.nbr_boxes = <int**> malloc(narrays*sizeof(int*))
        self.lengths = <int**> malloc(narrays*sizeof(int*))
        self.key_to_idx = <int**> malloc(narrays*sizeof(int*))

        cdef int i
        for i from 0<=i<narrays:
            self.pids[i] = NULL
            self.keys[i] = NULL
            self.cids[i] = NULL
            self.nbr_boxes[i] = NULL
            self.lengths[i] = NULL
            self.key_to_idx[i] = NULL

        self.current_pids = NULL
        self.current_cids_src = NULL
        self.current_cids_dst = NULL
        self.current_keys = NULL
        self.current_nbr_boxes = NULL
        self.current_lengths = NULL
        self.current_key_to_idx = NULL

    def __dealloc__(self):
        cdef int i
        for i from 0<=i<self.narrays:
            if self.pids[i] != NULL:
                free(self.pids[i])
            if self.keys[i] != NULL:
                free(self.keys[i])
            if self.key_to_idx[i] != NULL:
                free(self.key_to_idx[i])
            if self.cids[i] != NULL:
                free(self.cids[i])
            if self.nbr_boxes[i] != NULL:
                free(self.nbr_boxes[i])
            if self.lengths[i] != NULL:
                free(self.lengths[i])
        free(self.pids)
        free(self.keys)
        free(self.key_to_idx)
        free(self.cids)
        free(self.nbr_boxes)
        free(self.lengths)

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """
        NNPS.set_context(self, src_index, dst_index)
        self.current_pids = self.pids[src_index]
        self.current_keys = self.keys[src_index]
        self.current_key_to_idx = self.key_to_idx[src_index]
        self.current_cids_dst = self.cids[dst_index]
        self.current_cids_src = self.cids[src_index]
        self.current_nbr_boxes = self.nbr_boxes[src_index]
        self.current_lengths = self.lengths[src_index]

        self.dst = <NNPSParticleArrayWrapper> self.pa_wrappers[dst_index]
        self.src = <NNPSParticleArrayWrapper> self.pa_wrappers[src_index]

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        """Low level, high-performance non-gil method to find neighbors.
        This requires that `set_context()` be called beforehand.  This method
        does not reset the neighbors array before it appends the
        neighbors to it.

        """
        cdef double* dst_x_ptr = self.dst.x.data
        cdef double* dst_y_ptr = self.dst.y.data
        cdef double* dst_z_ptr = self.dst.z.data
        cdef double* dst_h_ptr = self.dst.h.data

        cdef double* src_x_ptr = self.src.x.data
        cdef double* src_y_ptr = self.src.y.data
        cdef double* src_z_ptr = self.src.z.data
        cdef double* src_h_ptr = self.src.h.data

        cdef double x = dst_x_ptr[d_idx]
        cdef double y = dst_y_ptr[d_idx]
        cdef double z = dst_z_ptr[d_idx]
        cdef double h = dst_h_ptr[d_idx]

        cdef int num_particles = self.src.x.length

        cdef uint32_t cid = self.current_cids_dst[d_idx]
        cdef uint32_t cid_nbr

        cdef unsigned int* s_gid = self.src.gid.data
        cdef int orig_length = nbrs.length

        cdef int c_x, c_y, c_z
        cdef double* xmin = self.xmin.data
        cdef int i, j

        find_cell_id_raw(
                x - xmin[0],
                y - xmin[1],
                z - xmin[2],
                self.h_sub,
                &c_x, &c_y, &c_z
                )

        cdef double xij2 = 0
        cdef double hi2 = self.radius_scale2*h*h
        cdef double hj2 = 0

        cdef int start_idx, curr_idx
        cdef uint32_t n, idx
        cdef uint64_t key
        cdef int length

        for i from 0<=i<self.mask_len:
            start_idx = self.current_nbr_boxes[self.mask_len * cid + i]

            if start_idx < 0:
                break

            key = self.current_keys[start_idx]
            idx = self.current_pids[start_idx]
            cid_nbr = self.current_cids_src[idx]
            length = self.current_lengths[cid_nbr]

            for j from 0<=j<length:
                idx = self.current_pids[start_idx + j]

                hj2 = self.radius_scale2*src_h_ptr[idx]*src_h_ptr[idx]

                xij2 = norm2(
                    src_x_ptr[idx] - x,
                    src_y_ptr[idx] - y,
                    src_z_ptr[idx] - z
                    )

                if (xij2 < hi2) or (xij2 < hj2):
                    nbrs.c_append(idx)

        if self.sort_gids:
            self._sort_neighbors(
                &nbrs.data[orig_length], nbrs.length - orig_length, s_gid
            )

    cpdef get_nearest_particles_no_cache(self, int src_index, int dst_index,
            size_t d_idx, UIntArray nbrs, bint prealloc):
        """Find nearest neighbors for particle id 'd_idx' without cache

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        d_idx: size_t
            Index of the query point in the destination particle array

        nbrs: UIntArray
            Array to be populated by nearest neighbors of 'd_idx'

        """
        self.set_context(src_index, dst_index)

        if prealloc:
            nbrs.length = 0
        else:
            nbrs.c_reset()

        self.find_nearest_neighbors(d_idx, nbrs)

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices):
        indices.reset()
        cdef NNPSParticleArrayWrapper pa_wrapper = self.pa_wrappers[pa_index]
        cdef int num_particles = pa_wrapper.get_number_of_particles()

        cdef uint32_t* current_pids = self.pids[pa_index]

        cdef int j
        for j from 0<=j<num_particles:
            indices.c_append(<long>current_pids[j])

    cdef int fill_array(self, NNPSParticleArrayWrapper pa_wrapper, int pa_index,
            uint32_t* current_pids, uint64_t* current_keys,
            uint32_t* current_cids, uint32_t curr_cid):

        cdef double* x_ptr = pa_wrapper.x.data
        cdef double* y_ptr = pa_wrapper.y.data
        cdef double* z_ptr = pa_wrapper.z.data

        cdef int curr_num_particles = pa_wrapper.get_number_of_particles()

        cdef double* xmin = self.xmin.data

        cdef int c_x, c_y, c_z

        cdef int i, n
        for i from 0<=i<curr_num_particles:
            find_cell_id_raw(
                    x_ptr[i] - xmin[0],
                    y_ptr[i] - xmin[1],
                    z_ptr[i] - xmin[2],
                    self.h_sub,
                    &c_x, &c_y, &c_z
                    )
            current_pids[i] = i
            current_keys[i] = get_key(c_x, c_y, c_z)

        cdef CompareSortWrapper sort_wrapper = \
                CompareSortWrapper(current_pids, current_keys, curr_num_particles)

        sort_wrapper.compare_sort()

        ################################################################

        current_key_to_idx = self.key_to_idx[pa_index]

        key = current_keys[0]
        current_key_to_idx[key] = 0;

        for i from 0<i<curr_num_particles:
            key = current_keys[i]
            if key != current_keys[i-1]:
                current_key_to_idx[key] = i;

        ################################################################

        cdef uint32_t found_cid = UINT_MAX
        cdef int num_particles
        cdef uint64_t* found_ptr = NULL
        cdef int found_idx

        cdef uint32_t pid

        cdef uint32_t* iter_cids
        cdef uint64_t* iter_keys
        cdef int* iter_key_to_idx

        cdef uint32_t curr_pid = current_pids[0]
        for j from 0<=j<pa_index:
            iter_cids = self.cids[j]
            iter_keys = self.keys[j]
            iter_key_to_idx = self.key_to_idx[j]
            iter_pids = self.pids[j]
            num_particles = (<NNPSParticleArrayWrapper> \
                    self.pa_wrappers[j]).get_number_of_particles()
            found_idx = iter_key_to_idx[current_keys[0]]

            if found_idx != -1:
                pid = iter_pids[found_idx]
                found_cid = iter_cids[pid]
                break

        if found_cid == UINT_MAX:
            found_cid = curr_cid
            curr_cid += 1

        current_cids[curr_pid] = found_cid

        for i from 0<i<curr_num_particles:
            curr_pid = current_pids[i]
            if(current_keys[i] != current_keys[i-1]):
                found_ptr = NULL
                found_cid = UINT_MAX

                for j from 0<=j<pa_index:
                    iter_cids = self.cids[j]
                    iter_keys = self.keys[j]
                    iter_key_to_idx = self.key_to_idx[j]
                    iter_pids = self.pids[j]
                    num_particles = (<NNPSParticleArrayWrapper> \
                            self.pa_wrappers[j]).get_number_of_particles()
                    found_idx = iter_key_to_idx[current_keys[i]]

                    if found_idx != -1:
                        pid = iter_pids[found_idx]
                        found_cid = iter_cids[pid]
                        break

                if found_cid == UINT_MAX:
                    found_cid = curr_cid
                    curr_cid += 1

            current_cids[curr_pid] = found_cid

        return curr_cid

    cpdef np.ndarray get_nbr_boxes(self, pa_index, cid):
        cdef IntArray nbr_boxes_arr = IntArray()
        cdef int* current_nbr_boxes = self.nbr_boxes[pa_index]
        nbr_boxes_arr.c_set_view(&current_nbr_boxes[self.mask_len*cid],
                self.mask_len)
        return nbr_boxes_arr.get_npy_array()

    cpdef np.ndarray get_keys(self, pa_index):
        cdef uint64_t* current_keys = self.keys[pa_index]
        cdef int num_particles = (<NNPSParticleArrayWrapper> \
                self.pa_wrappers[pa_index]).get_number_of_particles()
        keys = np.empty(num_particles)
        for i from 0<=i<num_particles:
            keys[i] = current_keys[i]
        return keys

    cpdef np.ndarray get_cids(self, pa_index):
        cdef UIntArray cids = UIntArray()
        cdef int num_particles = (<NNPSParticleArrayWrapper> \
                self.pa_wrappers[pa_index]).get_number_of_particles()
        cids.c_set_view(self.cids[pa_index], num_particles)
        return cids.get_npy_array()

    cpdef np.ndarray get_pids(self, pa_index):
        cdef UIntArray pids = UIntArray()
        cdef int num_particles = (<NNPSParticleArrayWrapper> \
                self.pa_wrappers[pa_index]).get_number_of_particles()
        pids.c_set_view(self.pids[pa_index], num_particles)
        return pids.get_npy_array()

    cdef void _fill_nbr_boxes(self):
        cdef int i, j, k
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int num_particles

        cdef int* current_nbr_boxes
        cdef uint32_t* current_cids
        cdef uint32_t* current_pids
        cdef uint64_t* current_keys
        cdef int* current_key_to_idx
        cdef int found_idx

        cdef double* x_ptr
        cdef double* y_ptr
        cdef double* z_ptr
        cdef double* h_ptr

        cdef int c_x, c_y, c_z
        cdef int num_boxes

        cdef int n = 0
        cdef uint32_t cid, pid

        cdef int found_indices[27]

        cdef uint64_t key

        for i from 0<=i<self.narrays:
            n = 0
            pa_wrapper = self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()
            x_ptr = pa_wrapper.x.data
            y_ptr = pa_wrapper.y.data
            z_ptr = pa_wrapper.z.data
            h_ptr = pa_wrapper.h.data
            xmin = self.xmin.data

            current_keys = self.keys[i]
            current_key_to_idx = self.key_to_idx[i]
            current_cids = self.cids[i]
            current_pids = self.pids[i]
            current_nbr_boxes = self.nbr_boxes[i]
            current_lengths = self.lengths[i]

            # init self.nbr_boxes to -1
            for j from 0<=j<self.mask_len * self.max_cid:
                current_nbr_boxes[j] = -1

            for j from 0<=j<self.max_cid:
                current_lengths[j] = 1

            pid = current_pids[0]
            cid = current_cids[pid]

            find_cell_id_raw(
                x_ptr[pid] - xmin[0],
                y_ptr[pid] - xmin[1],
                z_ptr[pid] - xmin[2],
                self.h_sub,
                &c_x, &c_y, &c_z
                )

            num_boxes = self._neighbor_boxes(c_x, c_y, c_z, current_key_to_idx,
                    num_particles, found_indices)

            for k from 0<=k<num_boxes:
                found_idx = found_indices[k]
                current_nbr_boxes[self.mask_len*cid + n] = found_idx
                n += 1

            # nbrs of all cids in particle array i
            for j from 0<j<num_particles:
                key = current_keys[j]
                pid = current_pids[j]
                cid = current_cids[pid]
                n = 0

                if key != current_keys[j-1]:
                    find_cell_id_raw(
                        x_ptr[pid] - xmin[0],
                        y_ptr[pid] - xmin[1],
                        z_ptr[pid] - xmin[2],
                        self.h_sub,
                        &c_x, &c_y, &c_z
                        )

                    num_boxes = self._neighbor_boxes(c_x, c_y, c_z, current_key_to_idx,
                            num_particles, found_indices)

                    for k from 0<=k<num_boxes:
                        found_idx = found_indices[k]
                        current_nbr_boxes[self.mask_len*cid + n] = found_idx
                        n += 1
                else:
                    current_lengths[cid] += 1

    cdef inline int get_idx(self, uint64_t key, int* key_to_idx) nogil:
        return -1 if key >= self.max_key else key_to_idx[key]

    cdef inline int _neighbor_boxes(self, int i, int j, int k,
            int* current_key_to_idx, int num_particles,
            int* found_indices) nogil:
        cdef int length = 0
        cdef int p, q, r
        cdef uint64_t key
        cdef int found_idx
        for p from -1<=p<2:
            for q from -1<=q<2:
                for r from -1<=r<2:
                    if i+r>=0 and j+q>=0 and k+p>=0:
                        key = get_key(i+r, j+q, k+p)
                        found_idx = self.get_idx(key, current_key_to_idx)
                        if found_idx != -1:
                            found_indices[length] = found_idx
                            length += 1
        return length

    @cython.cdivision(True)
    cpdef _refresh(self):
        cdef NNPSParticleArrayWrapper pa_wrapper

        cdef int i, j, num_particles
        cdef int c_x, c_y, c_z

        cdef double* xmax = self.xmax.data
        cdef double* xmin = self.xmin.data

        cdef uint32_t max_cid = 0

        cdef uint32_t* current_pids
        cdef uint64_t* current_keys
        cdef int* current_key_to_idx
        cdef uint32_t* current_cids

        self.h_sub = self.cell_size / self.H

        find_cell_id_raw(
                xmax[0] - xmin[0],
                xmax[1] - xmin[1],
                xmax[2] - xmin[2],
                self.h_sub,
                &c_x, &c_y, &c_z
                )

        cdef uint64_t max_key = 1 + get_key(c_x, c_y, c_z)
        self.max_key = max_key

        for i from 0<=i<self.narrays:
            if self.pids[i] != NULL:
                free(self.pids[i])
            if self.keys[i] != NULL:
                free(self.keys[i])
            if self.key_to_idx[i] != NULL:
                free(self.key_to_idx[i])
            if self.cids[i] != NULL:
                free(self.cids[i])

            pa_wrapper = <NNPSParticleArrayWrapper> self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()

            self.pids[i] = <uint32_t*> malloc(num_particles*sizeof(uint32_t))
            self.keys[i] = <uint64_t*> malloc(num_particles*sizeof(uint64_t))
            self.key_to_idx[i] = <int*> malloc(max_key*sizeof(int))
            self.cids[i] = <uint32_t*> malloc(num_particles*sizeof(uint32_t))

            current_pids = self.pids[i]
            current_cids = self.cids[i]
            current_keys = self.keys[i]
            current_key_to_idx = self.key_to_idx[i]

            for j from 0<=j<max_key:
                current_key_to_idx[j] = -1

            max_cid = self.fill_array(pa_wrapper, i, current_pids,
                    current_keys, current_cids, max_cid)

        for i from 0<=i<self.narrays:
            if self.nbr_boxes[i] != NULL:
                free(self.nbr_boxes[i])
            if self.lengths[i] != NULL:
                free(self.lengths[i])
            self.nbr_boxes[i] = <int*> malloc(self.mask_len * max_cid * sizeof(int))
            self.lengths[i] = <int*> malloc(max_cid*sizeof(int))

        self.max_cid = max_cid # this is max_cid + 1

        self._fill_nbr_boxes()

    @cython.cdivision(True)
    cpdef _bin(self, int pa_index, UIntArray indices):
        pass

cdef class ExtendedZOrderNNPS(ZOrderNNPS):

    """Find nearest neighbors using Z-Order space filling curve"""

    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int H=3,
            bint asymmetric=False):
        ZOrderNNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids, H=H, asymmetric=asymmetric
        )

    def __cinit__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int H=3,
            bint asymmetric=False):

        narrays = len(particles)

        self.hmax = <double**> malloc(narrays*sizeof(double*))
        self.current_hmax = NULL

        cdef int i
        for i from 0<=i<narrays:
            self.hmax[i] = NULL

    def __dealloc__(self):
        cdef int i
        for i from 0<=i<self.narrays:
            if self.hmax[i] != NULL:
                free(self.hmax[i])
        free(self.hmax)

    cdef inline int _h_mask_exact(self, int* x, int* y, int* z) nogil:
        cdef int length = 0
        cdef int s, t, u

        for s from -self.H<=s<=self.H:
            for t from -self.H<=t<=self.H:
                for u from -self.H<=u<=self.H:

                    x[length] = s
                    y[length] = t
                    z[length] = u
                    length += 1

        return length

    cdef int _neighbor_boxes_func(self, int i, int j, int k,
            int* current_key_to_idx, uint32_t* current_cids,
            double* current_hmax, int num_particles,
            int* found_indices, double h):
        if self.asymmetric:
            return self._neighbor_boxes_asym(i, j, k, current_key_to_idx,
                    current_cids, current_hmax, num_particles,
                    found_indices, h)
        else:
            return self._neighbor_boxes_sym(i, j, k, current_key_to_idx,
                    current_cids, current_hmax, num_particles,
                    found_indices, h)

    cdef int _neighbor_boxes_asym(self, int i, int j, int k,
            int* current_key_to_idx, uint32_t* current_cids,
            double* current_hmax, int num_particles,
            int* found_indices, double h) nogil:
        cdef int length = 0

        cdef uint64_t key
        cdef int found_idx

        cdef int x_temp, y_temp, z_temp

        cdef int s, t, u

        for s from -self.H<=s<=self.H:
            for t from -self.H<=t<=self.H:
                for u from -self.H<=u<=self.H:

                    x_temp = i + u
                    y_temp = j + t
                    z_temp = k + s

                    if x_temp >= 0 and y_temp >= 0 and z_temp >= 0:
                        key = get_key(x_temp, y_temp, z_temp)
                        found_idx = self.get_idx(key, current_key_to_idx)

                        if found_idx == -1:
                            continue

                        found_indices[length] = found_idx
                        length += 1

        return length

    @cython.cdivision(True)
    cdef int _neighbor_boxes_sym(self, int i, int j, int k,
            int* current_key_to_idx, uint32_t* current_cids,
            double* current_hmax, int num_particles,
            int* found_indices, double h) nogil:
        cdef int length = 0

        cdef uint64_t key

        cdef int x_temp, y_temp, z_temp

        cdef double h_local
        cdef int H

        cdef int s, t, u

        for s from -self.H<=s<=self.H:
            for t from -self.H<=t<=self.H:
                for u from -self.H<=u<=self.H:

                    x_temp = i + u
                    y_temp = j + t
                    z_temp = k + s

                    if x_temp >= 0 and y_temp >= 0 and z_temp >= 0:
                        key = get_key(x_temp, y_temp, z_temp)
                        found_idx = self.get_idx(key, current_key_to_idx)

                        if found_idx == -1:
                            continue

                        cid = current_cids[found_idx]

                        h_local = self.radius_scale * fmax(current_hmax[cid], h)
                        H = <int> ceil(h_local / self.h_sub)

                        if abs(u) <= H and abs(t) <= H and abs(s) <= H:
                            found_indices[length] = found_idx
                            length += 1

        return length

    cdef void _fill_nbr_boxes(self):
        cdef int i, j, k
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int num_particles

        cdef int* current_nbr_boxes
        cdef uint32_t* current_cids
        cdef uint32_t* current_pids
        cdef uint64_t* current_keys
        cdef int* current_key_to_idx
        cdef double* current_hmax
        cdef int found_idx

        cdef double* x_ptr
        cdef double* y_ptr
        cdef double* z_ptr
        cdef double* h_ptr

        cdef int c_x, c_y, c_z
        cdef int num_boxes

        cdef int n = 0
        cdef uint32_t cid, pid

        cdef int* found_indices = <int*> malloc(self.mask_len * \
                sizeof(int))
        cdef uint64_t key

        for i from 0<=i<self.narrays:
            n = 0
            pa_wrapper = self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()
            x_ptr = pa_wrapper.x.data
            y_ptr = pa_wrapper.y.data
            z_ptr = pa_wrapper.z.data
            h_ptr = pa_wrapper.h.data
            xmin = self.xmin.data

            current_keys = self.keys[i]
            current_key_to_idx = self.key_to_idx[i]
            current_cids = self.cids[i]
            current_pids = self.pids[i]
            current_nbr_boxes = self.nbr_boxes[i]
            current_lengths = self.lengths[i]
            current_hmax = self.hmax[i]

            # init self.nbr_boxes to -1
            for j from 0<=j<self.mask_len * self.max_cid:
                current_nbr_boxes[j] = -1

            for j from 0<=j<self.max_cid:
                current_lengths[j] = 1
                current_hmax[j] = 0

            pid = current_pids[0]
            cid = current_cids[pid]

            current_hmax[cid] = h_ptr[pid]

            for j from 0<j<num_particles:
                key = current_keys[j]
                pid = current_pids[j]
                cid = current_cids[pid]

                if key != current_keys[j-1]:
                    current_hmax[cid] = h_ptr[pid]
                else:
                    current_hmax[cid] = fmax(current_hmax[cid], h_ptr[pid])

            pid = current_pids[0]
            cid = current_cids[pid]

            find_cell_id_raw(
                x_ptr[pid] - xmin[0],
                y_ptr[pid] - xmin[1],
                z_ptr[pid] - xmin[2],
                self.h_sub,
                &c_x, &c_y, &c_z
                )

            num_boxes = self._neighbor_boxes_func(c_x, c_y, c_z,
                    current_key_to_idx, current_cids, current_hmax,
                    num_particles, found_indices, current_hmax[cid])

            for k from 0<=k<num_boxes:
                found_idx = found_indices[k]
                current_nbr_boxes[self.mask_len*cid + n] = found_idx
                n += 1

            # nbrs of all cids in particle array i
            for j from 0<j<num_particles:
                key = current_keys[j]
                pid = current_pids[j]
                cid = current_cids[pid]
                n = 0

                if key != current_keys[j-1]:
                    find_cell_id_raw(
                        x_ptr[pid] - xmin[0],
                        y_ptr[pid] - xmin[1],
                        z_ptr[pid] - xmin[2],
                        self.h_sub,
                        &c_x, &c_y, &c_z
                        )

                    num_boxes = self._neighbor_boxes_func(c_x, c_y, c_z,
                            current_key_to_idx, current_cids, current_hmax,
                            num_particles, found_indices, current_hmax[cid])

                    for k from 0<=k<num_boxes:
                        found_idx = found_indices[k]
                        current_nbr_boxes[self.mask_len*cid + n] = found_idx
                        n += 1
                else:
                    current_lengths[cid] += 1

        free(found_indices)

    @cython.cdivision(True)
    cpdef _refresh(self):
        cdef NNPSParticleArrayWrapper pa_wrapper

        cdef int i, j, num_particles
        cdef int c_x, c_y, c_z

        cdef double* xmax = self.xmax.data
        cdef double* xmin = self.xmin.data

        cdef uint32_t max_cid = 0

        cdef uint32_t* current_pids
        cdef uint64_t* current_keys
        cdef int* current_key_to_idx
        cdef uint32_t* current_cids

        self.h_sub = self.cell_size / self.H

        find_cell_id_raw(
                xmax[0] - xmin[0],
                xmax[1] - xmin[1],
                xmax[2] - xmin[2],
                self.h_sub,
                &c_x, &c_y, &c_z
                )

        cdef uint64_t max_key = 1 + get_key(c_x, c_y, c_z)
        self.max_key = max_key

        for i from 0<=i<self.narrays:
            if self.pids[i] != NULL:
                free(self.pids[i])
            if self.keys[i] != NULL:
                free(self.keys[i])
            if self.key_to_idx[i] != NULL:
                free(self.key_to_idx[i])
            if self.cids[i] != NULL:
                free(self.cids[i])

            pa_wrapper = <NNPSParticleArrayWrapper> self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()

            self.pids[i] = <uint32_t*> malloc(num_particles*sizeof(uint32_t))
            self.keys[i] = <uint64_t*> malloc(num_particles*sizeof(uint64_t))
            self.key_to_idx[i] = <int*> malloc(max_key*sizeof(int))
            self.cids[i] = <uint32_t*> malloc(num_particles*sizeof(uint32_t))

            current_pids = self.pids[i]
            current_cids = self.cids[i]
            current_keys = self.keys[i]
            current_key_to_idx = self.key_to_idx[i]

            for j from 0<=j<max_key:
                current_key_to_idx[j] = -1

            max_cid = self.fill_array(pa_wrapper, i, current_pids,
                    current_keys, current_cids, max_cid)

        for i from 0<=i<self.narrays:
            if self.nbr_boxes[i] != NULL:
                free(self.nbr_boxes[i])
            if self.lengths[i] != NULL:
                free(self.lengths[i])
            if self.hmax[i] != NULL:
                free(self.hmax[i])
            self.nbr_boxes[i] = <int*> malloc(self.mask_len * max_cid * sizeof(int))
            self.lengths[i] = <int*> malloc(max_cid*sizeof(int))
            self.hmax[i] = <double*> malloc(max_cid*sizeof(double))

        self.max_cid = max_cid # this is max_cid + 1

        self._fill_nbr_boxes()

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """
        ZOrderNNPS.set_context(self, src_index, dst_index)
        self.current_hmax = self.hmax[src_index]




"""Couette flow using the transport velocity formulation (30 seconds).
"""

import os
import numpy as np

# PySPH imports
from pysph.base.nnps import DomainManager
from pysph.base.utils import get_particle_array
from pysph.solver.utils import load
from pysph.solver.application import Application

from pysph.sph.scheme import TVFScheme

# domain and reference values
Re = 0.0125
d = 0.5; Ly = 2*d; Lx = 0.4*Ly
rho0 = 1.0; nu = 0.01

# upper wall velocity based on the Reynolds number and channel width
Vmax = nu*Re/(2*d)
print(Vmax)
c0 = 10*Vmax; p0 = c0*c0*rho0

# Numerical setup
dx = 0.05
ghost_extent = 5 * dx
hdx = 1.0

# adaptive time steps
h0 = hdx * dx
dt_cfl = 0.25 * h0/( c0 + Vmax )
dt_viscous = 0.125 * h0**2/nu
dt_force = 1.0

tf = 100.0
dt = min(dt_cfl, dt_viscous, dt_force)

class CouetteFlow(Application):
    def create_domain(self):
        return DomainManager(xmin=0, xmax=Lx, periodic_in_x=True)

    def create_particles(self):
        _x = np.arange( dx/2, Lx, dx )

        # create the fluid particles
        _y = np.arange( dx/2, Ly, dx )

        x, y = np.meshgrid(_x, _y); fx = x.ravel(); fy = y.ravel()

        # create the channel particles at the top
        _y = np.arange(Ly+dx/2, Ly+dx/2+ghost_extent, dx)
        x, y = np.meshgrid(_x, _y); tx = x.ravel(); ty = y.ravel()

        # create the channel particles at the bottom
        _y = np.arange(-dx/2, -dx/2-ghost_extent, -dx)
        x, y = np.meshgrid(_x, _y); bx = x.ravel(); by = y.ravel()

        # concatenate the top and bottom arrays
        cx = np.concatenate( (tx, bx) )
        cy = np.concatenate( (ty, by) )

        # create the arrays
        channel = get_particle_array(
            name='channel', x=cx, y=cy, rho=rho0*np.ones_like(cx)
        )
        fluid = get_particle_array(
            name='fluid', x=fx, y=fy, rho=rho0*np.ones_like(fx)
        )

        print("Couette flow :: Re = %g, nfluid = %d, nchannel=%d, dt = %g"%(
            Re, fluid.get_number_of_particles(),
            channel.get_number_of_particles(), dt))

        self.scheme.setup_properties([fluid, channel])

        # setup the particle properties
        volume = dx * dx

        # mass is set to get the reference density of rho0
        fluid.m[:] = volume * rho0
        channel.m[:] = volume * rho0

        # volume is set as dx^2
        fluid.V[:] = 1./volume
        channel.V[:] = 1./volume

        # smoothing lengths
        fluid.h[:] = hdx * dx
        channel.h[:] = hdx * dx

        # channel velocity on upper portion
        indices = np.where(channel.y > d)[0]
        channel.u[indices] = Vmax

        # return the particle list
        return [fluid, channel]

    def create_scheme(self):
        s = TVFScheme(
            ['fluid'], ['channel'], dim=2, rho0=rho0, c0=c0, nu=nu,
            p0=p0, pb=p0, h0=dx*hdx
        )
        s.configure_solver(tf=tf, dt=dt)
        return s

    def post_process(self, info_fname):
        info = self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')

        y_ex, u_ex, y, u = self._plot_u_vs_y()
        t, ke = self._plot_ke_history()
        res = os.path.join(self.output_dir, "results.npz")
        np.savez(res, t=t, ke=ke, y_ex=y_ex, u_ex=u_ex, y=y, u=u)

    def _plot_ke_history(self):
        from pysph.tools.pprocess import get_ke_history
        from matplotlib import pyplot as plt
        t, ke = get_ke_history(self.output_files, 'fluid')
        plt.clf()
        plt.plot(t, ke)
        plt.xlabel('t'); plt.ylabel('Kinetic energy')
        fig = os.path.join(self.output_dir, "ke_history.png")
        plt.savefig(fig, dpi=300)
        return t, ke

    def _plot_u_vs_y(self):
        files = self.output_files
        # take the last solution data
        fname = files[-1]
        data = load(fname)
        tf = data['solver_data']['t']
        fluid = data['arrays']['fluid']
        yp = fluid.y.copy()
        up = fluid.u.copy()

        # exact parabolic profile for the u-velocity
        ye = np.linspace(0, 1, 101)
        ue = Vmax*ye/Ly
        from matplotlib import pyplot as plt
        plt.clf()
        plt.plot(ye, ue, label="exact")
        plt.plot(yp, up, 'ko', fillstyle='none', label="computed")
        plt.xlabel('y'); plt.ylabel('u')
        plt.legend()
        plt.title('Velocity profile at %s'%tf)
        fig = os.path.join(self.output_dir, "comparison.png")
        plt.savefig(fig, dpi=300)
        return ye, ue, yp, up


if __name__ == '__main__':
    app = CouetteFlow()
    app.run()
    app.post_process(app.info_filename)



"""Taylor Green vortex flow (5 minutes).
"""

import os
import numpy as np
from numpy import pi, sin, cos, exp

from pysph.base.nnps import DomainManager
from pysph.base.utils import get_particle_array
from pysph.base.kernels import QuinticSpline
from pysph.solver.application import Application

from pysph.sph.equation import Group, Equation
from pysph.sph.scheme import TVFScheme, WCSPHScheme, SchemeChooser
from pysph.sph.wc.edac import ComputeAveragePressure, EDACScheme
from pysph.sph.iisph import IISPHScheme

from pysph.sph.wc.kernel_correction import (
    GradientCorrectionPreStep, GradientCorrection,
    MixedKernelCorrectionPreStep, MixedGradientCorrection
)
from pysph.sph.wc.crksph import CRKSPHPreStep, CRKSPH, CRKSPHScheme
from pysph.sph.wc.gtvf import GTVFScheme
from pysph.sph.wc.pcisph import PCISPHScheme
from pysph.sph.wc.shift import ShiftPositions
from pysph.sph.isph.sisph import SISPHScheme
from pysph.sph.isph.isph import ISPHScheme

#
# domain and constants
L = 1.0
U = 1.0
rho0 = 1.0
c0 = 10 * U
p0 = c0**2 * rho0


def m4p(x=0.0):
    """From the paper by Chaniotis et al. (JCP 2002).
    """
    if x < 0.0:
        return 0.0
    elif x < 1.0:
        return 1.0 - 0.5*x*x*(5.0 - 3.0*x)
    elif x < 2.0:
        return (1 - x)*(2 - x)*(2 - x)*0.5
    else:
        return 0.0


class M4(Equation):
    '''An equation to be used for remeshing.
    '''

    def initialize(self, d_idx, d_prop):
        d_prop[d_idx] = 0.0

    def _get_helpers_(self):
        return [m4p]

    def loop(self, s_idx, d_idx, s_temp_prop, d_prop, d_h, XIJ):
        xij = abs(XIJ[0]/d_h[d_idx])
        yij = abs(XIJ[1]/d_h[d_idx])
        d_prop[d_idx] += m4p(xij)*m4p(yij)*s_temp_prop[s_idx]


def exact_solution(U, b, t, x, y):
    factor = U * exp(b*t)

    u = -cos(2*pi*x) * sin(2*pi*y)
    v = sin(2*pi*x) * cos(2*pi*y)
    p = -0.25 * (cos(4*pi*x) + cos(4*pi*y))

    return factor * u, factor * v, factor * factor * p


class TaylorGreen(Application):

    def add_user_options(self, group):
        group.add_argument(
            "--init", action="store", type=str, default=None,
            help="Initialize particle positions from given file."
        )
        group.add_argument(
            "--perturb", action="store", type=float, dest="perturb", default=0,
            help="Random perturbation of initial particles as a fraction "
            "of dx (setting it to zero disables it, the default)."
        )
        group.add_argument(
            "--nx", action="store", type=int, dest="nx", default=50,
            help="Number of points along x direction. (default 50)"
        )
        group.add_argument(
            "--re", action="store", type=float, dest="re", default=100,
            help="Reynolds number (defaults to 100)."
        )
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.0,
            help="Ratio h/dx."
        )
        group.add_argument(
            "--pb-factor", action="store", type=float, dest="pb_factor",
            default=1.0,
            help="Use fraction of the background pressure (default: 1.0)."
        )
        corrections = ['', 'mixed', 'gradient', 'crksph']
        group.add_argument(
            "--kernel-correction", action="store", type=str,
            dest='kernel_correction',
            default='', help="Type of Kernel Correction", choices=corrections
        )
        group.add_argument(
            "--remesh", action="store", type=int, dest="remesh", default=0,
            help="Remeshing frequency (setting it to zero disables it)."
        )
        remesh_types = ['m4', 'sph']
        group.add_argument(
            "--remesh-eq", action="store", type=str, dest="remesh_eq",
            default='m4', choices=remesh_types,
            help="Remeshing strategy to use."
        )
        group.add_argument(
            "--shift-freq", action="store", type=int, dest="shift_freq",
            default=0,
            help="Particle position shift frequency.(set zero to disable)."
        )
        shift_types = ['simple', 'fickian']
        group.add_argument(
            "--shift-kind", action="store", type=str, dest="shift_kind",
            default='simple', choices=shift_types,
            help="Use of fickian shift in positions."
        )
        group.add_argument(
            "--shift-parameter", action="store", type=float,
            dest="shift_parameter", default=None,
            help="Constant used in shift, range for 'simple' is 0.01-0.1"
            "range 'fickian' is 1-10."
        )
        group.add_argument(
            "--shift-correct-vel", action="store_true",
            dest="correct_vel", default=False,
            help="Correct velocities after shifting (defaults to false)."
        )

    def consume_user_options(self):
        nx = self.options.nx
        re = self.options.re

        self.nu = nu = U * L / re

        self.dx = dx = L / nx
        self.volume = dx * dx
        self.hdx = self.options.hdx

        h0 = self.hdx * self.dx
        if self.options.scheme.endswith('isph'):
            dt_cfl = 0.25 * h0 / U
        else:
            dt_cfl = 0.25 * h0 / (c0 + U)
        dt_viscous = 0.125 * h0**2 / nu
        dt_force = 0.25 * 1.0

        self.dt = min(dt_cfl, dt_viscous, dt_force)
        self.tf = 2.0
        self.kernel_correction = self.options.kernel_correction

    def configure_scheme(self):
        scheme = self.scheme
        h0 = self.hdx * self.dx
        pfreq = 100
        kernel = QuinticSpline(dim=2)
        if self.options.scheme == 'tvf':
            scheme.configure(pb=self.options.pb_factor * p0, nu=self.nu, h0=h0)
        elif self.options.scheme == 'wcsph':
            scheme.configure(hdx=self.hdx, nu=self.nu, h0=h0)
        elif self.options.scheme == 'edac':
            scheme.configure(h=h0, nu=self.nu, pb=self.options.pb_factor * p0)
        elif self.options.scheme.endswith('isph'):
            pfreq = 10
            scheme.configure(nu=self.nu)
        elif self.options.scheme == 'crksph':
            scheme.configure(h0=h0, nu=self.nu)
        elif self.options.scheme == 'gtvf':
            scheme.configure(pref=p0, nu=self.nu, h0=h0)
        scheme.configure_solver(kernel=kernel, tf=self.tf, dt=self.dt,
                                pfreq=pfreq)

    def create_scheme(self):
        h0 = None
        hdx = None
        wcsph = WCSPHScheme(
            ['fluid'], [], dim=2, rho0=rho0, c0=c0, h0=h0,
            hdx=hdx, nu=None, gamma=7.0, alpha=0.0, beta=0.0
        )
        tvf = TVFScheme(
            ['fluid'], [], dim=2, rho0=rho0, c0=c0, nu=None,
            p0=p0, pb=None, h0=h0
        )
        edac = EDACScheme(
            ['fluid'], [], dim=2, rho0=rho0, c0=c0, nu=None,
            pb=p0, h=h0
        )
        iisph = IISPHScheme(
            fluids=['fluid'], solids=[], dim=2, nu=None,
            rho0=rho0, has_ghosts=True
        )
        crksph = CRKSPHScheme(
            fluids=['fluid'], dim=2, nu=None,
            rho0=rho0, h0=h0, c0=c0, p0=0.0
        )
        gtvf = GTVFScheme(
            fluids=['fluid'], solids=[], dim=2, rho0=rho0, c0=c0,
            nu=None, h0=None, pref=None
        )
        pcisph = PCISPHScheme(
            fluids=['fluid'], dim=2, rho0=rho0, nu=None
        )
        sisph = SISPHScheme(
            fluids=['fluid'], solids=[], dim=2, nu=None, rho0=rho0,
            c0=c0, alpha=0.0, has_ghosts=True, pref=p0,
            rho_cutoff=0.2, internal_flow=True, gtvf=True
        )
        isph = ISPHScheme(
            fluids=['fluid'], solids=[], dim=2, nu=None, rho0=rho0, c0=c0,
            alpha=0.0
        )
        s = SchemeChooser(
            default='tvf', wcsph=wcsph, tvf=tvf, edac=edac, iisph=iisph,
            crksph=crksph, gtvf=gtvf, pcisph=pcisph, sisph=sisph, isph=isph
        )
        return s

    def create_equations(self):
        eqns = self.scheme.get_equations()
        # This tolerance needs to be fixed.
        tol = 0.5
        if self.kernel_correction == 'gradient':
            cls1 = GradientCorrectionPreStep
            cls2 = GradientCorrection
        elif self.kernel_correction == 'mixed':
            cls1 = MixedKernelCorrectionPreStep
            cls2 = MixedGradientCorrection
        elif self.kernel_correction == 'crksph':
            cls1 = CRKSPHPreStep
            cls2 = CRKSPH

        if self.kernel_correction:
            g1 = Group(equations=[cls1('fluid', ['fluid'], dim=2)])
            eq2 = cls2(dest='fluid', sources=['fluid'], dim=2, tol=tol)

            if self.options.scheme == 'wcsph':
                eqns.insert(1, g1)
                eqns[2].equations.insert(0, eq2)
            elif self.options.scheme == 'tvf':
                eqns[1].equations.append(g1.equations[0])
                eqns[2].equations.insert(0, eq2)
            elif self.options.scheme == 'edac':
                eqns.insert(1, g1)
                eqns[2].equations.insert(0, eq2)

        return eqns

    def create_domain(self):
        return DomainManager(
            xmin=0, xmax=L, ymin=0, ymax=L, periodic_in_x=True,
            periodic_in_y=True
        )

    def create_fluid(self):
        # create the particles
        dx = self.dx
        _x = np.arange(dx / 2, L, dx)
        x, y = np.meshgrid(_x, _x)
        if self.options.init is not None:
            fname = self.options.init
            from pysph.solver.utils import load
            data = load(fname)
            _f = data['arrays']['fluid']
            x, y = _f.x.copy(), _f.y.copy()

        if self.options.perturb > 0:
            np.random.seed(1)
            factor = dx * self.options.perturb
            x += np.random.random(x.shape) * factor
            y += np.random.random(x.shape) * factor

        # Initialize
        m = self.volume * rho0
        h = self.hdx * dx
        re = self.options.re
        b = -8.0*pi*pi / re
        u0, v0, p0 = exact_solution(U=U, b=b, t=0, x=x, y=y)
        color0 = cos(2*pi*x) * cos(4*pi*y)

        # create the arrays
        fluid = get_particle_array(name='fluid', x=x, y=y, m=m, h=h, u=u0,
                                   v=v0, rho=rho0, p=p0, color=color0)
        return fluid

    def create_particles(self):
        fluid = self.create_fluid()

        self.scheme.setup_properties([fluid])

        print("Taylor green vortex problem :: nfluid = %d, dt = %g" % (
            fluid.get_number_of_particles(), self.dt))

        # volume is set as dx^2
        if self.options.scheme == 'sisph':
            nfp = fluid.get_number_of_particles()
            fluid.gid[:] = np.arange(nfp)
            fluid.add_output_arrays(['gid'])
        if self.options.scheme == 'tvf':
            fluid.V[:] = 1. / self.volume
        if self.options.scheme == 'iisph':
            # These are needed to update the ghost particle properties.
            nfp = fluid.get_number_of_particles()
            fluid.orig_idx[:] = np.arange(nfp)
            fluid.add_output_arrays(['orig_idx'])

        corr = self.kernel_correction
        if corr in ['mixed', 'crksph']:
            fluid.add_property('cwij')
        if corr == 'mixed' or corr == 'gradient':
            fluid.add_property('m_mat', stride=9)
            fluid.add_property('dw_gamma', stride=3)
        elif corr == 'crksph':
            fluid.add_property('ai')
            fluid.add_property('gradbi', stride=4)
            for prop in ['gradai', 'bi']:
                fluid.add_property(prop, stride=2)

        if self.options.shift_freq > 0:
            fluid.add_constant('vmax', [0.0])
            fluid.add_property('dpos', stride=3)
            fluid.add_property('gradv', stride=9)

        if self.options.scheme == 'isph':
            gid = np.arange(fluid.get_number_of_particles(real=False))
            fluid.add_property('gid')
            fluid.gid[:] = gid[:]
            fluid.add_property('dpos', stride=3)
            fluid.add_property('gradv', stride=9)

        return [fluid]

    def create_tools(self):
        tools = []
        options = self.options
        if options.remesh > 0:
            if options.remesh_eq == 'm4':
                equations = [M4(dest='interpolate', sources=['fluid'])]
            else:
                equations = None
            from pysph.solver.tools import SimpleRemesher
            if options.scheme == 'wcsph' or options.scheme == 'crksph':
                props = ['u', 'v', 'au', 'av', 'ax', 'ay', 'arho']
            elif options.scheme == 'pcisph':
                props = ['u', 'v', 'p']
            elif options.scheme == 'tvf':
                props = ['u', 'v', 'uhat', 'vhat',
                         'au', 'av', 'auhat', 'avhat']
            elif options.scheme == 'edac':
                if 'uhat' in self.particles[0].properties:
                    props = ['u', 'v', 'uhat', 'vhat', 'p',
                             'au', 'av', 'auhat', 'avhat', 'ap']
                else:
                    props = ['u', 'v', 'p', 'au', 'av', 'ax', 'ay', 'ap']
            elif options.scheme == 'iisph' or options.scheme == 'isph':
                # The accelerations are not really needed since the current
                # stepper is a single stage stepper.
                props = ['u', 'v', 'p']
            elif options.scheme == 'gtvf':
                props = [
                    'uhat', 'vhat', 'what', 'rho0', 'rhodiv', 'p0',
                    'auhat', 'avhat', 'awhat', 'arho', 'arho0'
                ]

            remesher = SimpleRemesher(
                self, 'fluid', props=props,
                freq=self.options.remesh, equations=equations
            )
            tools.append(remesher)

        if options.shift_freq > 0:
            shift = ShiftPositions(
                self, 'fluid', freq=self.options.shift_freq,
                shift_kind=self.options.shift_kind,
                correct_velocity=self.options.correct_vel,
                parameter=self.options.shift_parameter
            )
            tools.append(shift)

        return tools

    # The following are all related to post-processing.
    def _get_post_process_props(self, array):
        """Return x, y, m, u, v, p.
        """
        if 'pavg' not in array.properties or \
           'pavg' not in array.output_property_arrays:
            self._add_extra_props(array)
            sph_eval = self._get_sph_evaluator(array)
            sph_eval.update_particle_arrays([array])
            sph_eval.evaluate()

        x, y, m, u, v, p, pavg = array.get(
            'x', 'y', 'm', 'u', 'v', 'p', 'pavg'
        )
        return x, y, m, u, v, p - pavg

    def _add_extra_props(self, array):
        extra = ['pavg', 'nnbr']
        for prop in extra:
            if prop not in array.properties:
                array.add_property(prop)
        array.add_output_arrays(extra)

    def _get_sph_evaluator(self, array):
        if not hasattr(self, '_sph_eval'):
            from pysph.tools.sph_evaluator import SPHEvaluator
            equations = [
                ComputeAveragePressure(dest='fluid', sources=['fluid'])
            ]
            dm = self.create_domain()
            sph_eval = SPHEvaluator(
                arrays=[array], equations=equations, dim=2,
                kernel=QuinticSpline(dim=2), domain_manager=dm
            )
            self._sph_eval = sph_eval
        return self._sph_eval

    def post_process(self, info_fname):
        info = self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output
        decay_rate = -8.0 * np.pi**2 / self.options.re

        files = self.output_files
        t, ke, ke_ex, decay, linf, l1, p_l1 = [], [], [], [], [], [], []
        for sd, array in iter_output(files, 'fluid'):
            _t = sd['t']
            t.append(_t)
            x, y, m, u, v, p = self._get_post_process_props(array)
            u_e, v_e, p_e = exact_solution(U, decay_rate, _t, x, y)
            vmag2 = u**2 + v**2
            vmag = np.sqrt(vmag2)
            ke.append(0.5 * np.sum(m * vmag2))
            vmag2_e = u_e**2 + v_e**2
            vmag_e = np.sqrt(vmag2_e)
            ke_ex.append(0.5 * np.sum(m * vmag2_e))

            vmag_max = vmag.max()
            decay.append(vmag_max)
            theoretical_max = U * np.exp(decay_rate * _t)
            linf.append(abs((vmag_max - theoretical_max) / theoretical_max))

            l1_err = np.average(np.abs(vmag - vmag_e))
            avg_vmag_e = np.average(np.abs(vmag_e))
            # scale the error by the maximum velocity.
            l1.append(l1_err / avg_vmag_e)

            p_e_max = np.abs(p_e).max()
            p_error = np.average(np.abs(p - p_e)) / p_e_max
            p_l1.append(p_error)

        t, ke, ke_ex, decay, l1, linf, p_l1 = list(map(
            np.asarray, (t, ke, ke_ex, decay, l1, linf, p_l1))
        )
        decay_ex = U * np.exp(decay_rate * t)
        fname = os.path.join(self.output_dir, 'results.npz')
        np.savez(
            fname, t=t, ke=ke, ke_ex=ke_ex, decay=decay, linf=linf, l1=l1,
            p_l1=p_l1, decay_ex=decay_ex
        )

        import matplotlib
        matplotlib.use('Agg')

        from matplotlib import pyplot as plt
        plt.clf()
        plt.semilogy(t, decay_ex, label="exact")
        plt.semilogy(t, decay, label="computed")
        plt.xlabel('t')
        plt.ylabel('max velocity')
        plt.legend()
        fig = os.path.join(self.output_dir, "decay.png")
        plt.savefig(fig, dpi=300)

        plt.clf()
        plt.plot(t, linf)
        plt.xlabel('t')
        plt.ylabel(r'$L_\infty$ error')
        fig = os.path.join(self.output_dir, "linf_error.png")
        plt.savefig(fig, dpi=300)

        plt.clf()
        plt.plot(t, l1, label="error")
        plt.xlabel('t')
        plt.ylabel(r'$L_1$ error')
        fig = os.path.join(self.output_dir, "l1_error.png")
        plt.savefig(fig, dpi=300)

        plt.clf()
        plt.plot(t, p_l1, label="error")
        plt.xlabel('t')
        plt.ylabel(r'$L_1$ error for $p$')
        fig = os.path.join(self.output_dir, "p_l1_error.png")
        plt.savefig(fig, dpi=300)

    def customize_output(self):
        self._mayavi_config('''
        b = particle_arrays['fluid']
        b.scalar = 'vmag'
        ''')


if __name__ == '__main__':
    app = TaylorGreen()
    app.run()
    app.post_process(app.info_filename)



from numpy import pi, sin, cos, exp

from pysph.examples.taylor_green import TaylorGreen, exact_solution
from pysph.base.utils import get_particle_array

# domain and constants
L = 1.0
U = 1.0
rho0 = 1.0
c0 = 10 * U
p0 = c0**2 * rho0


class TGPacked(TaylorGreen):
    def _get_packed_points(self):
        '''
        returns
        xs, ys, zs, xf, yf, zf
        '''
        from pysph.tools.geometry import get_packed_periodic_packed_particles
        folder = self.output_dir
        dx = self.dx
        return get_packed_periodic_packed_particles(
            self.add_user_options, folder, dx, L=L, B=L)

    def create_fluid(self):
        # create the particles
        xs, ys, zs, xf, yf, zf = self._get_packed_points()
        x, y = xf, yf
        if self.options.init is not None:
            fname = self.options.init
            from pysph.solver.utils import load
            data = load(fname)
            _f = data['arrays']['fluid']
            x, y = _f.x.copy(), _f.y.copy()

        # Initialize
        dx = self.dx
        m = self.volume * rho0
        h = self.hdx * dx
        re = self.options.re
        b = -8.0 * pi * pi / re
        u0, v0, p0 = exact_solution(U=U, b=b, t=0, x=x, y=y)
        color0 = cos(2 * pi * x) * cos(4 * pi * y)

        # create the arrays
        fluid = get_particle_array(
            name='fluid', x=x, y=y, m=m, h=h, u=u0, v=v0, rho=rho0, p=p0,
            color=color0)
        return fluid


if __name__ == '__main__':
    app = TGPacked()
    app.run()
    app.post_process(app.info_filename)



#cython: boundscheck=False

#
# Eigen decomposition code for symmetric 3x3 matrices, some code taken
# from the public domain Java Matrix library JAMA

from libc.math cimport sqrt, cos, acos, sin, atan2, M_PI
from libc.string cimport memcpy

from numpy.linalg import eigh
cimport numpy
import numpy

cdef extern:
    double fabs(double) nogil

# this is cython substitute for const values
cdef enum:
    n=3

cdef double EPS = numpy.finfo(float).eps

cdef inline double MAX(double a, double b) nogil:
    return a if a>b else b

cdef inline double SQR(double a) nogil:
    return a*a

cdef inline double hypot2(double x, double y) nogil:
    return sqrt(x*x+y*y)


cdef double det(double a[3][3]) nogil:
    '''Determinant of symmetrix matrix
    '''
    return (a[0][0]*a[1][1]*a[2][2] + 2*a[1][2]*a[0][2]*a[0][1] -
            a[0][0]*a[1][2]*a[1][2] - a[1][1]*a[0][2]*a[0][2] -
            a[2][2]*a[0][1]*a[0][1])

cpdef double py_det(double[:,:] m):
    '''Determinant of symmetrix matrix
    '''
    return det(<double(*)[3]>&m[0][0])

# d,s are diagonal and off-diagonal elements of a symmetric 3x3 matrix
# d:11,22,33; s:23,13,12
# d:00,11,22; s:12,02,01

cdef void get_eigenvalues(double a[3][3], double *result) nogil:
    '''Compute the eigenvalues of symmetric matrix a and return in
    result array.
    '''
    cdef double m = (a[0][0]+ a[1][1]+a[2][2])/3.0
    cdef double K[3][3]
    memcpy(&K[0][0], &a[0][0], sizeof(double)*9)
    K[0][0], K[1][1], K[2][2] = a[0][0] - m, a[1][1] - m, a[2][2] - m
    cdef double q = det(K)*0.5
    cdef double p = 0
    p += K[0][0]*K[0][0] + 2*K[1][2]*K[1][2]
    p += K[1][1]*K[1][1] + 2*K[0][2]*K[0][2]
    p += K[2][2]*K[2][2] + 2*K[0][1]*K[0][1]
    p /= 6.0
    cdef double pi = M_PI
    cdef double phi = 0.5*pi
    cdef double tmp = p**3 - q**2

    if q == 0.0 and p == 0.0:
        # singular zero matrix
        result[0] = result[1] = result[2] = m
        return
    elif tmp < 0.0 or fabs(tmp) < EPS: # eliminate roundoff error
        phi = 0
    else:
        phi = atan2(sqrt(tmp), q)/3.0
    if phi == 0 and q < 0:
        phi = pi

    result[0] = m + 2*sqrt(p)*cos(phi)
    result[1] = m - sqrt(p)*(cos(phi) + sqrt(3)*sin(phi))
    result[2] = m - sqrt(p)*(cos(phi) - sqrt(3)*sin(phi))

cpdef py_get_eigenvalues(double[:,:] m):
    '''Return the eigenvalues of symmetric matrix.
    '''
    res = numpy.empty(3, float)
    cdef double[:] _res  = res
    get_eigenvalues(<double(*)[3]>&m[0][0], &_res[0])
    return res


##############################################################################
cdef void get_eigenvector_np(double A[n][n], double r, double *res):
    ''' eigenvector of symmetric matrix for given eigenvalue `r` using numpy
    '''
    cdef numpy.ndarray[ndim=2, dtype=numpy.float64_t] mat=numpy.empty((3,3)), evec
    cdef numpy.ndarray[ndim=1, dtype=numpy.float64_t] evals
    cdef double[:,:] _mat = mat
    cdef int i, j
    for i in range(3):
        for j in range(3):
            _mat[i,j] = A[i][j]
    evals, evec = eigh(mat)
    cdef int idx=0
    cdef double di = fabs(evals[0]-r)
    if fabs(evals[1]-r) < di:
        idx = 1
    if fabs(evals[2]-r) < di:
        idx = 2
    for i in range(3):
        res[i] = evec[idx,i]

cdef void get_eigenvector(double A[n][n], double r, double *res):
    ''' get eigenvector of symmetric 3x3 matrix for given eigenvalue `r`

    uses a fast method to get eigenvectors with a fallback to using numpy
    '''
    res[0] = A[0][1]*A[1][2] - A[0][2]*(A[1][1]-r) # a_01 * a_12 - a_02 * (a_11 - r)
    res[1] = A[0][1]*A[0][2] - A[1][2]*(A[0][0]-r) # a_01 * a_02 - a_12 * (a_00 - r)
    res[2] = (A[0][0]-r)*(A[1][1]-r) - A[0][1]*A[0][1] # (a_00 - r) * (a_11 - r) - a_01^2
    cdef double norm = sqrt(SQR(res[0]) + SQR(res[1]) + SQR(res[2]))

    if norm *1e7 <= fabs(r):
        # its a zero, let numpy get the answer
        get_eigenvector_np(A, r, res)
    else:
        res[0] /= norm
        res[1] /= norm
        res[2] /= norm

cpdef py_get_eigenvector(double[:,:] A, double r):
    ''' get eigenvector of a symmetric matrix for given eigenvalue `r` '''
    d = numpy.empty(3, dtype=float)
    cdef double[:] _d = d
    get_eigenvector(<double(*)[n]>&A[0,0], r, &_d[0])
    return d

cdef void get_eigenvec_from_val(double A[n][n], double *R, double *e):
    cdef int i, j
    cdef double res[3]
    for i in range(3):
        get_eigenvector(A, e[i], &res[0])
        for j in range(3):
            R[j*3+i] = res[j]


cdef bint _nearly_diagonal(double A[n][n]) nogil:
    return (
        (SQR(A[0][0]) + SQR(A[1][1]) + SQR(A[2][2])) >
        1e8*(SQR(A[0][1]) + SQR(A[0][2]) + SQR(A[1][2]))
    )

cdef void get_eigenvalvec(double A[n][n], double *R, double *e):
    '''Get the eigenvalues and eigenvectors of symmetric 3x3 matrix.

    A is the input 3x3 matrix.
    R is the output eigen matrix
    e are the output eigenvalues
    '''
    cdef bint use_iter = False
    cdef int i,j
    if A[0][1] == A[0][2] == A[1][2] == 0.0:
        # diagonal matrix.
        e[0] = A[0][0]
        e[1] = A[1][1]
        e[2] = A[2][2]
        for i in range(3):
            for j in range(3):
                R[i*3+j] = (i==j)
        return

    # FIXME: implement fast version
    get_eigenvalues(A, e)
    if e[0] != e[1] and e[1] != e[2] and e[0] != e[2]:
        # no repeated eigenvalues
        use_iter = True
    if _nearly_diagonal(A):
        # nearly diagonal matrix
        use_iter = True
    if not use_iter:
        get_eigenvec_from_val(A, R, e)
    else:
        eigen_decomposition(
            A, <double(*)[n]>&R[0], &e[0]
        )

def py_get_eigenvalvec(double[:,:] A):
    v = numpy.empty((3,3), dtype=float)
    d = numpy.empty(3, dtype=float)
    cdef double[:,:] _v = v
    cdef double[:] _d = d
    get_eigenvalvec(<double(*)[n]>&A[0,0], &_v[0,0], &_d[0])
    return d, v


##############################################################################

cdef void transform(double A[3][3], double P[3][3], double res[3][3]) nogil:
    '''Compute the transformation P.T*A*P and add it into res.
    '''
    cdef int i, j, k, l
    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    res[i][j] += P[k][i]*A[k][l]*P[l][j] # P.T*A*P

cdef void transform_diag(double *A, double P[3][3],
                         double res[3][3]) nogil:
    '''Compute the transformation P.T*A*P and add it into res.

    A is diagonal and contains the diagonal entries alone.
    '''
    cdef int i, j, k
    for i in range(3):
        for j in range(3):
            for k in range(3):
                res[i][j] += P[k][i]*A[k]*P[k][j] # P.T*A*P

cdef void transform_diag_inv(double *A, double P[3][3],
                             double res[3][3]) nogil:
    '''Compute the transformation P*A*P.T and set it into res.
    A is diagonal and contains just the diagonal entries.
    '''
    cdef int i, j, k
    for i in range(3):
        for j in range(3):
            res[i][j] = 0.0

    for i in range(3):
        for j in range(3):
            for k in range(3):
                res[i][j] += P[i][k]*A[k]*P[j][k] # P*A*P.T

def py_transform(double[:,:] A, double[:,:] P):
    res = numpy.zeros((3,3), dtype=float)
    cdef double[:,:] _res = res
    transform(
        <double(*)[3]>&A[0][0], <double(*)[3]>&P[0][0],
        <double(*)[3]>&_res[0][0]
    )
    return res

def py_transform_diag(double[:] A, double[:,:] P):
    res = numpy.zeros((3,3), dtype=float)
    cdef double[:,:] _res = res
    transform_diag(
        &A[0], <double(*)[3]>&P[0][0], <double(*)[3]>&_res[0][0]
    )
    return res

def py_transform_diag_inv(double[:] A, double[:,:] P):
    res = numpy.empty((3,3), dtype=float)
    cdef double[:,:] _res = res
    transform_diag_inv(
        &A[0], <double(*)[3]>&P[0][0], <double(*)[3]>&_res[0][0]
    )
    return res


cdef double * tred2(double V[n][n], double *d, double *e) nogil:
    '''Symmetric Householder reduction to tridiagonal form

    This is derived from the Algol procedures tred2 by
    Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
    Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
    Fortran subroutine in EISPACK.

    d contains the diagonal elements of the tridiagonal matrix.

    e contains the subdiagonal elements of the tridiagonal matrix in its
    last n-1 positions.  e[0] is set to zero.
    '''

    cdef:
        double scale, f, g, h, hh
        int i, j, k

    for j in range(n):
        d[j] = V[n-1][j]

    # Householder reduction to tridiagonal form.

    for i in range(n-1,0,-1):

        # Scale to avoid under/overflow.
        scale = 0.0
        h = 0.0
        for k in range(i):
            scale += fabs(d[k])

        if (scale == 0.0):
            e[i] = d[i-1]
            for j in range(i):
                d[j] = V[i-1][j]
                V[i][j] = 0.0
                V[j][i] = 0.0

        else:
            # Generate Householder vector.
            for k in range(i):
                d[k] /= scale
                h += d[k] * d[k]

            f = d[i-1]
            g = sqrt(h)
            if f > 0:
                g = -g

            e[i] = scale * g
            h = h - f * g
            d[i-1] = f - g
            for j in range(i):
                e[j] = 0.0

        # Apply similarity transformation to remaining columns.

            for j in range(i):
                f = d[j]
                V[j][i] = f
                g = e[j] + V[j][j] * f
                for k in range(j+1, i):
                    g += V[k][j] * d[k]
                    e[k] += V[k][j] * f

                e[j] = g

            f = 0.0
            for j in range(i):
                e[j] /= h
                f += e[j] * d[j]

            hh = f / (h + h)
            for j in range(i):
                e[j] -= hh * d[j]

            for j in range(i):
                f = d[j]
                g = e[j]
                for k in range(j,i):
                    V[k][j] -= (f * e[k] + g * d[k])

                d[j] = V[i-1][j];
                V[i][j] = 0.0;

        d[i] = h

    # Accumulate transformations.

    for i in range(n-1):
        V[n-1][i] = V[i][i]
        V[i][i] = 1.0
        h = d[i+1]
        if h != 0.0:
            for k in range(i+1):
                d[k] = V[k][i+1] / h

            for j in range(i+1):
                g = 0.0
                for k in range(i+1):
                    g += V[k][i+1] * V[k][j]

                for k in range(i+1):
                    V[k][j] -= g * d[k]


        for k in range(i+1):
            V[k][i+1] = 0.0

    for j in range(n):
        d[j] = V[n-1][j]
        V[n-1][j] = 0.0

    V[n-1][n-1] = 1.0
    e[0] = 0.0

    return d


cdef void tql2(double V[n][n], double *d, double *e) nogil:
    '''Symmetric tridiagonal QL algo for eigendecomposition

    This is derived from the Algol procedures tql2, by
    Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
    Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
    Fortran subroutine in EISPACK.

    d contains the eigenvalues in ascending order.  if an error exit is
    made, the eigenvalues are correct but unordered for indices
    1,2,...,ierr-1.

    e has been destroyed.
    '''

    cdef:
        double f, tst1, eps, g, h, p, r, dl1, c, c2, c3, el1, s, s2
        int i, j, k, l, m, iter
        bint cont

    for i in range(1, n):
        e[i-1] = e[i]

    e[n-1] = 0.0

    f = 0.0
    tst1 = 0.0
    eps = 2.0**-52.0
    for l in range(n):

        # Find small subdiagonal element
        tst1 = MAX(tst1,fabs(d[l]) + fabs(e[l]))
        m = l
        while m < n:
            if fabs(e[m]) <= eps*tst1:
                break
            m += 1

        # If m == l, d[l] is an eigenvalue,
        # otherwise, iterate.
        if m > l:
            iter = 0
            cont = True
            while cont:
                iter = iter + 1        # (Could check iteration count here.)

                # Compute implicit shift
                g = d[l]
                p = (d[l+1] - g) / (2.0 * e[l])
                r = hypot2(p,1.0)
                if p < 0:
                    r = -r

                d[l] = e[l] / (p + r)
                d[l+1] = e[l] * (p + r)
                dl1 = d[l+1]
                h = g - d[l]
                for i in range(l+2,n):
                    d[i] -= h

                f += h

                # Implicit QL transformation.
                p = d[m]
                c = 1.0
                c2 = c
                c3 = c
                el1 = e[l+1]
                s = 0.0
                s2 = 0.0
                for i in range(m-1,l-1,-1):
                    c3 = c2
                    c2 = c
                    s2 = s
                    g = c * e[i]
                    h = c * p
                    r = hypot2(p,e[i])
                    e[i+1] = s * r
                    s = e[i] / r
                    c = p / r
                    p = c * d[i] - s * g
                    d[i+1] = h + s * (c * g + s * d[i])

                    # Accumulate transformation
                    for k in range(n):
                        h = V[k][i+1]
                        V[k][i+1] = s * V[k][i] + c * h
                        V[k][i] = c * V[k][i] - s * h

                p = -s * s2 * c3 * el1 * e[l] / dl1
                e[l] = s * p
                d[l] = c * p

                # Check for convergence
                cont = bool(fabs(e[l]) > eps*tst1)
        d[l] += f
        e[l] = 0.0

    # Sort eigenvalues and corresponding vectors.
    for i in range(n-1):
        k = i
        p = d[i]
        for j in range(i+1,n):
            if d[j] < p:
                k = j
                p = d[j]

        if k != i:
            d[k] = d[i]
            d[i] = p
            for j in range(n):
                p = V[j][i]
                V[j][i] = V[j][k]
                V[j][k] = p


cdef void zero_matrix_case(double V[n][n], double *d) nogil:
    cdef int i, j
    for i in range(3):
        d[i] = 0.0
        for j in range(3):
            V[i][j] = (i==j)

cdef void eigen_decomposition(double A[n][n], double V[n][n], double *d) nogil:
    '''Get eigenvalues and eigenvectors of matrix A.
    V is output eigenvectors and d are the eigenvalues.
    '''
    cdef double e[n]
    cdef int i, j
    # Scale the matrix, as if the matrix is tiny, floating point errors
    # creep up leading to zero division errors in tql2.  This is
    # specifically tested for with a tiny matrix.
    cdef double s = 0.0
    for i in range(n):
        for j in range(n):
            V[i][j] = A[i][j]
            s += fabs(V[i][j])

    if s == 0:
        zero_matrix_case(V, d)
    else:
        for i in range(n):
            for j in range(n):
                V[i][j] /= s

        d = tred2(V, d, &e[0])
        tql2(V, d, &e[0])
        for i in range(n):
            d[i] *= s


def py_eigen_decompose_eispack(double[:,:] a):
    v = numpy.empty((3,3), dtype=float)
    d = numpy.empty(3, dtype=float)
    cdef double[:,:] _v = v
    cdef double[:] _d = d
    eigen_decomposition(
        <double(*)[n]>&a[0,0], <double(*)[n]>&_v[0,0], &_d[0]
    )
    return d, v
